# October 2021, Lewis Gaul

"""
Minesweeper minefield API.

"""

__all__ = ("MinefieldBase",)

import abc
import logging
import random
from typing import Any, Generic, Iterable, List, Mapping, Optional, Set, TypeVar

from .board import BoardBase


logger = logging.getLogger(__name__)


C = TypeVar("C")
B = TypeVar("B", bound=BoardBase)


class MinefieldBase(Generic[C, B], metaclass=abc.ABCMeta):
    """Representation of a minesweeper minefield, generic over the coord type."""

    # At the most basic level, all the minefield needs to do is choose where mines
    # are going to be in the available cells. The simplest way to implement this
    # (and therefore easiest to make generic) is to pass in a list of coordinates
    # that can contain mines, and the minefield then just places mines in those
    # coordinates.
    # The responsibility of calculating properties of the minefield (e.g. 3bv,
    # final board, ...) can go in mode-specific implementations (e.g. subclasses).

    def __init__(
        self,
        all_coords: Iterable[C],
        *,
        mines: int,
        per_cell: int = 1,
    ):
        """
        :param all_coords:
            Iterable of coords where mines may lie.
        :param mines:
            The number of mines to randomly place.
        :param per_cell:
            Maximum number of mines per cell.
        :raise ValueError:
            If the number of mines is too high.
        """
        self.all_coords: Set[C] = set(all_coords)
        self.mines: int = mines
        self.per_cell: int = per_cell
        self.mine_coords: List[C] = []
        self._bbbv: Optional[int] = None
        self._completed_board: Optional[B] = None
        self._openings: Optional[List[List[C]]] = None
        self.populated: bool = False

        # Perform some checks on the args.
        if self.per_cell < 1:
            raise ValueError(
                f"Max mines per cell must be at least 1, got {self.per_cell}"
            )
        if self.mines < 0:
            raise ValueError(f"Number of mines must be positive, got {self.mines}")
        mine_spaces = len(self.all_coords) - 1
        if self.mines > mine_spaces * self.per_cell:
            raise ValueError(
                f"Number of mines too high: {self.mines} in {mine_spaces} "
                f"spaces with max {self.per_cell} per cell"
            )

    @classmethod
    def from_coords(
        cls,
        all_coords: Iterable[C],
        *,
        mine_coords: Iterable[C],
        per_cell: int = 1,
    ) -> "MinefieldBase":
        """
        :param all_coords:
            Iterable of coords where mines may lie.
        :param mine_coords:
            Iterable of coords containing mines.
        :param per_cell:
            Maximum number of mines per cell.
        :raise ValueError:
            If the number of mines is too high.
        """
        mine_coords = list(mine_coords)
        if any(mine_coords.count(c) > per_cell for c in mine_coords):
            raise ValueError(
                f"Max number of mines per cell is {per_cell}, too many in: "
                + ", ".join(
                    str(c) for c in mine_coords if mine_coords.count(c) > per_cell
                )
            )
        self = cls(all_coords, mines=len(mine_coords), per_cell=per_cell)
        self.mine_coords = mine_coords
        self.populated = True
        return self

    def __repr__(self):
        return f"<Minefield with {len(self.all_coords)} coords, {self.mines} mines>"

    def __getitem__(self, coord: C) -> int:
        if coord not in self.all_coords:
            raise IndexError(f"Invalid coord '{coord}'")
        return self.mine_coords.count(coord)

    def __eq__(self, other):
        if type(other) is not type(self):
            return False
        fields = ("all_coords", "mines", "mine_coords", "per_cell")
        return all(getattr(self, f) == getattr(other, f) for f in fields)

    @property
    def bbbv(self) -> int:
        if not self.populated:
            raise AttributeError("Uninitialised minefield has no 3bv")
        if self._bbbv is None:
            self._bbbv = self._calc_3bv()
        return self._bbbv

    @property
    def completed_board(self) -> B:
        if not self.populated:
            raise AttributeError("Uninitialised minefield has no openings")
        if self._completed_board is None:
            self._completed_board = self._calc_completed_board()
        return self._completed_board

    @property
    def openings(self) -> List[List[C]]:
        if not self.populated:
            raise AttributeError("Uninitialised minefield has no openings")
        if self._openings is None:
            self._openings = self._find_openings()
        return self._openings

    def populate(self, safe_coords: Optional[Iterable[C]] = None) -> None:
        """
        Randomly place mines in the available coordinates.

        :param safe_coords:
            Optional iterable of coords that should not contain a mine when
            filling the minefield.
        :return:
            A list of randomly chosen mine coords.
        :raise ValueError:
            If the number of mines is too high.
        """
        safe_coords = set(safe_coords) if safe_coords else None
        mine_spaces = len(self.all_coords) - (len(safe_coords) if safe_coords else 1)

        if self.mines > mine_spaces * self.per_cell:
            raise ValueError(
                f"Number of mines too high: {self.mines} in {mine_spaces} "
                f"spaces with max {self.per_cell} per cell"
            )

        # Get a list of coordinates which can have mines placed in them.
        # Make sure there is at least one safe cell.
        avble_set = set(self.all_coords)
        if safe_coords:
            avble_set -= safe_coords
        else:
            avble_set.remove(random.choice(tuple(avble_set)))

        avble_list = list(avble_set) * self.per_cell
        random.shuffle(avble_list)
        self.mine_coords = avble_list[: self.mines]
        self.populated = True
        logger.debug("Populated minefield with %s mines", len(self.mine_coords))

    @abc.abstractmethod
    def _calc_3bv(self) -> int:
        """Calculate the 3bv of the board."""
        raise NotImplementedError

    @abc.abstractmethod
    def _calc_completed_board(self) -> B:
        """
        Create the completed board with the flags and numbers that should be
        seen upon game completion.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def _find_openings(self) -> List[List[C]]:
        """Find the openings in the completed board."""
        raise NotImplementedError

    @abc.abstractmethod
    def to_json(self) -> Mapping[str, Any]:
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def from_json(cls, obj: Mapping[str, Any]) -> "MinefieldBase":
        raise NotImplementedError



# October 2021, Lewis Gaul

__all__ = ("MinefieldWidgetImpl",)

import logging
from typing import Dict

from PyQt5.QtGui import QPixmap
from PyQt5.QtWidgets import QGraphicsScene

from ... import api
from ...core.regular import Coord
from ...shared.types import CellContents, CellImageType
from ..state import State
from . import _base
from ._base import FlagAction, update_cell_images


logger = logging.getLogger(__name__)


class MinefieldWidgetImpl(_base.MinefieldWidgetImplBase):
    """
    Minefield widget for regular game.
    """

    def __init__(
        self, scene: QGraphicsScene, ctrlr: api.AbstractController, state: State
    ):
        super().__init__(scene, ctrlr, state)
        self._cell_images: Dict[CellContents, QPixmap] = {}
        self.update_cell_images()

    def set_cell_image(self, coord: Coord, state: CellContents) -> None:
        """Set the image of a cell on the scene."""
        try:
            img = self._cell_images[state]
        except KeyError:
            logger.error("Missing cell image for state: %s", state)
            return
        b = self._scene.addPixmap(img)
        b.setPos(coord.x * self.btn_size, coord.y * self.btn_size)

    def update_cell_images(self, img_type: CellImageType = CellImageType.ALL) -> None:
        update_cell_images(
            self._cell_images, self.btn_size, self._state.styles, img_type
        )

    def right_down_action(self, coord: Coord) -> FlagAction:
        """Perform a right click, returning the action that was taken."""
        self._ctrlr.flag_cell(coord)
        if self._ctrlr.board[coord] is CellContents.Unclicked:
            return FlagAction.UNFLAG
        else:
            return FlagAction.FLAG

    def right_drag_action(self, coord: Coord, action: FlagAction) -> None:
        """Perform a right drag click that matches the given action."""
        if action is FlagAction.FLAG:
            self._ctrlr.flag_cell(coord, flag_only=True)
        elif action is FlagAction.UNFLAG:
            self._ctrlr.remove_cell_flags(coord)
        else:
            assert False, f"Unexpected action {action!r}"



# June 2018, Lewis Gaul

"""
Types used throughout the project.

Exports
-------
.. class:: CellContents
    An ADT-like class providing cell contents types.

.. class:: CellImageType
    An enum of cell image types.

.. class:: Coord
    A base coordinate class.

.. class:: Difficulty
    An enum of board difficulties.

.. class:: FaceState
    An enum of possible face states.

.. class:: GameState
    An enum of states a game can be in.

.. class:: GameMode
    An enum of game modes.

.. class:: PathLike
    A more permissive version of the `os.PathLike` type alias.

.. class:: UIMode
    An enum of supported UI modes.

"""

__all__ = (
    "CellContents",
    "CellImageType",
    "Coord",
    "Difficulty",
    "FaceState",
    "GameMode",
    "GameState",
    "PathLike",
    "UIMode",
)

import abc
import enum
import functools
import os
from typing import Union


PathLike = Union[str, bytes, os.PathLike]


# ------------------------------------------------------------------------------
# Coordinate types
# ------------------------------------------------------------------------------


class Coord(metaclass=abc.ABCMeta):
    fields = ("x", "y")

    x: int
    y: int

    def __repr__(self):
        return "{}({})".format(
            type(self).__name__,
            ", ".join(f"{x}={getattr(self, x)}" for x in self.fields),
        )

    def __eq__(self, other):
        try:
            return all(getattr(self, x) == getattr(other, x) for x in self.fields)
        except AttributeError:
            return False

    def __hash__(self):
        return hash(tuple(getattr(self, x) for x in self.fields))

    def __lt__(self, other):
        if not isinstance(other, Coord):
            return NotImplemented
        return (self.x, self.y) < (other.x, other.y)


# ------------------------------------------------------------------------------
# Cell contents types
# ------------------------------------------------------------------------------


class _NumericCellContentsMixin:
    """
    A mixin for numeric cell contents types, allowing adding and subtracting integers.
    """

    char: str

    def __init__(self, num):
        if not isinstance(num, int):
            raise TypeError("Number should be an integer")
        self.num = num

    def __repr__(self):
        return self.char + str(self.num)

    def __add__(self, obj):
        if type(obj) is not int:
            raise TypeError("Can only add integers to cell contents types")
        else:
            return self.__class__(self.num + obj)

    def __sub__(self, obj):
        if type(obj) is not int:
            raise TypeError("Can only subtract integers from cell contents types")
        else:
            return self.__class__(self.num - obj)


class CellContents:
    """Abstract base class for contents of a minesweeper board cell."""

    char: str

    Unclicked = NotImplemented
    UnclickedSunken = NotImplemented
    Num = NotImplemented
    Mine = NotImplemented
    HitMine = NotImplemented
    Flag = NotImplemented
    WrongFlag = NotImplemented

    items = NotImplemented

    @functools.lru_cache(maxsize=None)
    def __new__(cls, *args):
        if cls is CellContents:
            raise TypeError("Base class should not be instantiated")
        return super().__new__(cls)

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return self.char

    @staticmethod
    def from_char(char: str) -> "CellContents":
        return NotImplemented  # Implemented below, after subclasses

    @staticmethod
    def from_str(string: str) -> "CellContents":
        return NotImplemented  # Implemented below, after subclasses

    def is_type(self, item: "CellContents") -> bool:
        if item in [self.Unclicked, self.UnclickedSunken]:
            return self is item
        elif item in self.items:
            return type(self) is item
        else:
            raise ValueError(f"Unrecognised type {item!r}")

    def is_mine_type(self) -> bool:
        return False  # Overridden by subclasses as required


class _CellUnclicked(CellContents):
    """Unclicked cell on a minesweeper board."""

    char = "#"


class _CellUnclickedSunken(_CellUnclicked):
    """Unclicked sunken cell on a minesweeper board."""


class _CellNum(_NumericCellContentsMixin, CellContents):
    """Number shown in a cell on a minesweeper board."""

    char = ""

    def __init__(self, num):
        super().__init__(num)
        if num < 0:
            raise ValueError("Cell value cannot be negative")


class _CellMineType(_NumericCellContentsMixin, CellContents):
    """Abstract base class for the number of a mine type in a cell."""

    def __new__(cls, num=1):
        if cls is _CellMineType:
            raise TypeError(
                f"{type(cls)} should be used as a base class and not "
                "instantiated directly"
            )
        return super().__new__(cls, num)

    def __init__(self, num=1):
        super().__init__(num)
        if num < 1:
            raise ValueError("Mine-type cell contents must represent one or more mines")

    def is_mine_type(self) -> bool:
        return True


class _CellMine(_CellMineType):
    """Number of mines in a cell shown on a minesweeper board."""

    char = "M"


class _CellHitMine(_CellMineType):
    """Number of hit mines in a cell shown on a minesweeper board."""

    char = "!"


class _CellFlag(_CellMineType):
    """Number of flags in a cell shown on a minesweeper board."""

    char = "F"


class _CellWrongFlag(_CellFlag):
    """Number of incorrect flags in a cell shown on a minesweeper board."""

    char = "X"


# Make the base class act like an ADT, serving as the only external API.
CellContents.Unclicked = _CellUnclicked()
CellContents.UnclickedSunken = _CellUnclickedSunken()
CellContents.Num = _CellNum
CellContents.Mine = _CellMine
CellContents.HitMine = _CellHitMine
CellContents.Flag = _CellFlag
CellContents.WrongFlag = _CellWrongFlag

CellContents.items = [
    CellContents.Unclicked,
    CellContents.UnclickedSunken,
    CellContents.Num,
    CellContents.Mine,
    CellContents.HitMine,
    CellContents.Flag,
    CellContents.WrongFlag,
]


def _from_char(char: str) -> CellContents:
    """
    Get the class of mine-like cell contents using the character
    representation.

    :param char:
        The character representation of a cell contents type.
    :return:
        The cell contents enum item.
    """
    for item in CellContents.items:
        if item.char == char:
            return item


def _from_str(string: str) -> CellContents:
    if string.isnumeric():
        return CellContents.Num(int(string))
    elif len(string) == 2:
        char, num = string
        return CellContents.from_char(char)(int(num))
    elif string == CellContents.Unclicked.char:
        return CellContents.Unclicked
    else:
        raise ValueError(f"Unknown cell contents representation {string!r}")


CellContents.from_char = _from_char
CellContents.from_str = _from_str


# ------------------------------------------------------------------------------
# Game enums
# ------------------------------------------------------------------------------


class Difficulty(str, enum.Enum):
    """Enum of difficulty settings."""

    BEGINNER = "B"
    INTERMEDIATE = "I"
    EXPERT = "E"
    MASTER = "M"
    LUDICROUS = "L"
    CUSTOM = "C"

    @classmethod
    def from_str(cls, value: str) -> "Difficulty":
        """Create an instance from a string representation."""
        if value.upper() in [x.name for x in cls]:
            value = value[0].upper()
        elif value.upper() in [x.value for x in cls]:
            value = value.upper()
        return cls(value)


class GameState(str, enum.Enum):
    """Enum representing the state of a game."""

    READY = "READY"
    ACTIVE = "ACTIVE"
    WON = "WON"
    LOST = "LOST"

    def started(self) -> bool:
        return self is not self.READY

    def finished(self) -> bool:
        return self in [self.WON, self.LOST]


class GameMode(str, enum.Enum):
    """Minesweeper game mode."""

    REGULAR = "regular"
    SPLIT_CELL = "split-cell"

    @classmethod
    def from_str(cls, value: str) -> "Difficulty":
        """Create an instance from a string representation."""
        sanitised_value = value.upper().replace("-", "_")
        if sanitised_value in [x.name for x in cls]:
            return getattr(cls, sanitised_value)
        else:
            return cls(value)  # Raise standard exception


# ------------------------------------------------------------------------------
# GUI enums
# ------------------------------------------------------------------------------


class FaceState(enum.Enum):
    """Enum of the 'new game' button face states."""

    READY = "ready"
    ACTIVE = "active"
    WON = "won"
    LOST = "lost"


class CellImageType(enum.Flag):
    """Enum of cell image types."""

    BUTTONS = enum.auto()
    NUMBERS = enum.auto()
    MARKERS = enum.auto()
    ALL = BUTTONS | NUMBERS | MARKERS


class UIMode(enum.Enum):
    """Enum of UI modes."""

    GAME = enum.auto()
    CREATE = enum.auto()



# October 2021, Lewis Gaul

from typing import Iterable, List, Union

from ...shared import utils
from ...shared.types import CellContents, GameMode
from ..board import BoardBase
from .types import Coord


class Board(BoardBase):
    """A regular minesweeper board."""

    mode = GameMode.REGULAR

    def __init__(self, x_size: int, y_size: int):
        self._grid = utils.Grid(x_size, y_size, fill=CellContents.Unclicked)

    @classmethod
    def from_2d_array(cls, array: List[List[Union[str, int]]]) -> "Board":
        """
        Create an instance from a 2-dimensional array.

        Cell representations:
         - Numbers are represented by ints
         - Unclicked cells are represented by '#'
         - Flags are represented by e.g. 'F1'
         - Mines are represented by e.g. 'M1'
         - Wrong flags are represented by e.g. 'X1'
         - Hit mines are represented by e.g. '!1'
        """
        grid = utils.Grid.from_2d_array(array)
        board = cls(grid.x_size, grid.y_size)
        for c in grid.all_coords:
            if type(grid[c]) is int:
                board[c] = CellContents.Num(grid[c])
            elif type(grid[c]) is str and len(grid[c]) == 2:
                char, num = grid[c]
                board[c] = CellContents.from_char(char)(int(num))
            elif grid[c] != CellContents.Unclicked.char:
                raise ValueError(
                    f"Unknown cell contents representation in cell {c}: {grid[c]}"
                )
        return board

    def __repr__(self):
        return f"<{self.x_size}x{self.y_size} board>"

    def __str__(self):
        return self._grid.__str__(mapping={CellContents.Num(0): "."})

    def __eq__(self, other) -> bool:
        if not isinstance(other, Board):
            return False
        return self._grid == other._grid

    def __getitem__(self, coord: Coord) -> CellContents:
        return self._grid[coord]

    def __setitem__(self, coord: Coord, value: CellContents):
        if not isinstance(value, CellContents):
            raise TypeError("Board can only contain CellContents instances")
        self._grid[coord] = value

    def __contains__(self, coord: Coord) -> bool:
        return coord in self.all_coords

    @property
    def x_size(self):
        return self._grid.x_size

    @property
    def y_size(self):
        return self._grid.y_size

    @property
    def all_coords(self) -> List[Coord]:
        return [Coord(*c) for c in self._grid.all_coords]

    def fill(self, value: CellContents) -> None:
        for c in self.all_coords:
            self[c] = value

    def get_nbrs(self, coord: Coord, *, include_origin=False) -> Iterable[Coord]:
        return [
            Coord(*c) for c in self._grid.get_nbrs(coord, include_origin=include_origin)
        ]

    def get_coord_at(self, x: int, y: int) -> Coord:
        if x < 0 or x >= self.x_size or y < 0 or y >= self.y_size:
            raise ValueError(f"Position out of bounds: ({x}, {y})")
        return Coord(x, y)

    def reset(self):
        """Reset the board to the initial state."""
        for c in self.all_coords:
            self[c] = CellContents.Unclicked



# October 2021, Lewis Gaul

__all__ = (
    "Coord",
    "Minefield",
    "Board",
    "Game",
    "GameController",
    "CreateController",
    "mode",
)

from ...shared.types import GameMode
from .board import Board
from .controller import CreateController, GameController
from .game import Game
from .minefield import Minefield
from .types import Coord


mode = GameMode.REGULAR



# February 2022, Lewis Gaul

"""
MySQL highscores database interface.
"""

__all__ = ("MySQLDB",)

import logging
import os
from typing import Iterable, Optional, Tuple

import attr
import mysql.connector
import mysql.connector.cursor

from ..shared.types import Difficulty, GameMode
from .base import AbstractHighscoresDB, HighscoreStruct, SQLMixin


################################################################################
# WARNING: This module should not be used by the minegauler package directly as
#          it adds the mysql-connector dependency.
################################################################################


logger = logging.getLogger(__name__)


class MySQLDB(SQLMixin, AbstractHighscoresDB):
    """MySQL highscores database."""

    _USER = "admin"
    _HOST = "minegauler-highscores.cb4tvkuqujyi.eu-west-2.rds.amazonaws.com"
    _DB_NAME = "minegauler"

    _cached_conn: Optional[mysql.connector.MySQLConnection] = None

    def __init__(self):
        """
        :raise mysql.connector.Error:
            If connecting to the DB fails for any reason.
        """
        cls = type(self)
        if not cls._cached_conn:
            logger.info("Initialising connection to MySQL highscores DB")
            cls._cached_conn = mysql.connector.connect(
                user=self._USER,
                password=self._PASSWORD,
                host=self._HOST,
                database=self._DB_NAME,
            )
        self._conn = cls._cached_conn

    @property
    def conn(self) -> mysql.connector.MySQLConnection:
        return self._conn

    @property
    def _PASSWORD(self):
        return os.environ.get("MINEGAULER_MYSQL_DB_PASSWORD")

    def get_highscores(
        self,
        *,
        game_mode: Optional[GameMode] = None,
        difficulty: Optional[Difficulty] = None,
        per_cell: Optional[int] = None,
        drag_select: Optional[bool] = None,
        name: Optional[str] = None,
    ) -> Iterable[HighscoreStruct]:
        super().get_highscores(
            game_mode=game_mode,
            difficulty=difficulty,
            per_cell=per_cell,
            drag_select=drag_select,
            name=name,
        )
        if game_mode is None:
            modes = list(GameMode)
        else:
            modes = [game_mode]
        ret = []
        for mode in modes:
            cursor = self.execute(
                self._get_select_highscores_sql(
                    game_mode=mode,
                    difficulty=difficulty,
                    per_cell=per_cell,
                    drag_select=drag_select,
                    name=name,
                ),
                dictionary=True,
            )
            ret.extend(HighscoreStruct(game_mode=mode, **r) for r in cursor.fetchall())
        return ret

    def count_highscores(self) -> int:
        """Count the number of rows in the highscores table."""
        super().count_highscores()
        return next(self.execute(self._get_highscores_count_sql()))[0]

    def insert_highscores(self, highscores: Iterable[HighscoreStruct]) -> None:
        super().insert_highscores(highscores)
        orig_count = self.count_highscores()
        for mode in GameMode:
            mode_rows = [attr.astuple(h)[1:] for h in highscores if h.game_mode is mode]
            self.executemany(
                self._get_insert_highscore_sql(fmt="?", game_mode=mode),
                mode_rows,
                commit=True,
            )
        return self.count_highscores() - orig_count

    def execute(
        self, cmd: str, params: Tuple = (), *, commit=False, **cursor_args
    ) -> mysql.connector.cursor.MySQLCursor:
        return super().execute(cmd, params, commit=commit, **cursor_args)

    def executemany(
        self, cmd: str, params: Iterable[Tuple] = (), *, commit=False, **cursor_args
    ) -> mysql.connector.cursor.MySQLCursor:
        return super().executemany(cmd, params, commit=commit, **cursor_args)



import argparse
import logging
import os
import sys

from flask import Flask

from .routes import init_route_handling


logger = logging.getLogger(__name__)

app = Flask(__name__)


# ------------------------------------------------------------------------------
# Main
# ------------------------------------------------------------------------------


def _parse_args(argv):
    parser = argparse.ArgumentParser()
    parser.add_argument("--port", "-p", type=int, help="Override the default port")
    parser.add_argument("--dev", action="store_true", help="Run in development mode")
    return parser.parse_args(argv)


def main(argv):
    args = _parse_args(argv)

    logging.basicConfig(
        filename="bot-server.log",
        level=logging.DEBUG,
        format="%(asctime)s %(name)-12s %(levelname)-8s %(message)s",
    )

    init_route_handling(app)

    logger.info("Starting up")
    if args.dev:
        os.environ["FLASK_ENV"] = "development"
        app.run(debug=True, port=args.port)
    else:
        from waitress import serve

        serve(app, port=args.port)


main(sys.argv[1:])



# January 2022, Lewis Gaul

__all__ = (
    "BOARDS_DIR",
    "DATA_DIR",
    "FILES_DIR",
    "HIGHSCORES_FILE",
    "IMG_DIR",
    "ROOT_DIR",
    "SETTINGS_FILE",
)

import pathlib
import sys


if hasattr(sys, "frozen") and hasattr(sys, "_MEIPASS"):  # in pyinstaller EXE
    ROOT_DIR = pathlib.Path(__file__).parents[2] / "app"
else:
    ROOT_DIR = pathlib.Path(__file__).parent
SETTINGS_FILE: pathlib.Path = ROOT_DIR / "settings.cfg"
DATA_DIR: pathlib.Path = ROOT_DIR / "data"
HIGHSCORES_FILE: pathlib.Path = DATA_DIR / "highscores.db"
IMG_DIR: pathlib.Path = ROOT_DIR / "images"
FILES_DIR: pathlib.Path = ROOT_DIR / "files"
BOARDS_DIR = ROOT_DIR / "boards"



# February 2022, Lewis Gaul

"""
Base classes/implementations for highscore handling.

"""

__all__ = (
    "HighscoreSettingsStruct",
    "HighscoreStruct",
    "AbstractHighscoresDB",
    "SQLMixin",
)

import abc
import logging
import textwrap
from typing import Iterable, Mapping, Optional, Tuple

import attr

from ..shared.types import Difficulty, GameMode
from ..shared.utils import StructConstructorMixin


logger = logging.getLogger(__name__)


@attr.attrs(auto_attribs=True, frozen=True)
class HighscoreSettingsStruct(StructConstructorMixin):
    """A set of highscore settings."""

    game_mode: GameMode = attr.attrib(converter=GameMode.from_str)
    difficulty: Difficulty = attr.attrib(converter=Difficulty.from_str)
    per_cell: int
    drag_select: bool = attr.attrib(converter=bool)

    @classmethod
    def get_default(cls) -> "HighscoreSettingsStruct":
        return cls(GameMode.REGULAR, Difficulty.BEGINNER, 1, False)


@attr.attrs(auto_attribs=True, frozen=True)
class HighscoreStruct(HighscoreSettingsStruct):
    """A single highscore."""

    name: str
    timestamp: int
    elapsed: float
    bbbv: int
    bbbvps: float
    flagging: float


class AbstractHighscoresDB(abc.ABC):
    """Abstract base class for a highscores database."""

    @property
    @abc.abstractmethod
    def conn(self):
        """The active database connection."""
        return NotImplemented

    @staticmethod
    def extract_single_elem(cursor):
        """Extract a single element using a cursor."""
        return next(cursor)[0]

    @abc.abstractmethod
    def get_highscores(
        self,
        *,
        game_mode: Optional[GameMode] = None,
        difficulty: Optional[Difficulty] = None,
        per_cell: Optional[int] = None,
        drag_select: Optional[bool] = None,
        name: Optional[str] = None,
    ) -> Iterable[HighscoreStruct]:
        """Fetch highscores from the database using the given filters."""
        logger.debug("%s: Getting highscores", type(self).__name__)
        return NotImplemented

    @abc.abstractmethod
    def count_highscores(self) -> int:
        """Count the number of rows in the highscores table."""
        logger.debug("%s: Counting number of highscores in DB", type(self).__name__)
        return NotImplemented

    @abc.abstractmethod
    def insert_highscores(self, highscores: Iterable[HighscoreStruct]) -> int:
        """Insert highscores into the database."""
        logger.debug("%s: Inserting highscores into DB", type(self).__name__)
        return NotImplemented

    def execute(self, cmd: str, params: Tuple = (), *, commit=False, **cursor_args):
        """
        Execute a command on the database.

        :param cmd:
            The command to execute.
        :param params:
            Parameters to pass to the command.
        :param commit:
            Whether to do a commit after executing the command.
        :param cursor_args:
            Keyword arguments to pass on when creating the DB cursor.
        """
        cursor = self.conn.cursor(**cursor_args)
        logger.debug(
            "%s: Executing command %r with params: %s", type(self).__name__, cmd, params
        )
        cursor.execute(cmd, params)
        if commit:
            self.conn.commit()
        return cursor

    def executemany(
        self, cmd: str, params: Iterable[Tuple] = (), *, commit=False, **cursor_args
    ):
        """
        Execute multiple commands on the database.

        :param cmd:
            The commands to execute.
        :param params:
            Parameters to pass to the command.
        :param commit:
            Whether to do a commit after executing the command.
        :param cursor_args:
            Keyword arguments to pass on when creating the DB cursor.
        """
        cursor = self.conn.cursor(**cursor_args)
        logger.debug("%s: Executing commands:\n%r", type(self).__name__, cmd)
        cursor.executemany(cmd, params)
        if commit:
            self.conn.commit()
        return cursor


class SQLMixin:
    """A mixin for SQL-like highscores databases."""

    TABLES: Mapping[GameMode, str] = {m: m.name.lower() for m in GameMode}

    _table_fields = [f.name for f in attr.fields(HighscoreStruct)][1:]

    def _get_create_table_sql(self, table_name: str) -> str:
        return textwrap.dedent(
            f"""\
            CREATE TABLE IF NOT EXISTS {table_name} (
                difficulty VARCHAR(1) NOT NULL,
                per_cell INTEGER NOT NULL,
                drag_select INTEGER NOT NULL,
                name VARCHAR(20) NOT NULL,
                timestamp INTEGER NOT NULL,
                elapsed REAL NOT NULL,
                bbbv INTEGER NOT NULL,
                bbbvps REAL NOT NULL,
                flagging REAL NOT NULL,
                PRIMARY KEY(difficulty, name, timestamp)
            )"""
        )

    def _get_select_highscores_sql(
        self,
        *,
        game_mode: GameMode,
        difficulty: Optional[Difficulty] = None,
        per_cell: Optional[int] = None,
        drag_select: Optional[bool] = None,
        name: Optional[str] = None,
    ) -> str:
        """Get the SQL command to get/select highscores from a DB."""
        conditions = []
        if difficulty is not None:
            conditions.append(f"difficulty='{difficulty.value}'")
        if per_cell is not None:
            conditions.append(f"per_cell={per_cell}")
        if drag_select is not None:
            conditions.append(f"drag_select={drag_select:d}")
        if name is not None:
            conditions.append(f"LOWER(name)='{name.lower()}'")
        return "SELECT {fields} FROM {table} {where} ORDER BY elapsed ASC".format(
            fields=", ".join(self._table_fields),
            table=self.TABLES[game_mode],
            where="WHERE " + " AND ".join(conditions) if conditions else "",
        )

    def _get_insert_highscore_sql(self, fmt="%s", *, game_mode: GameMode) -> str:
        """Get the SQL command to insert a highscore into a DB."""
        return "INSERT OR IGNORE INTO {table} ({fields}) VALUES ({fmt_})".format(
            table=self.TABLES[game_mode],
            fields=", ".join(self._table_fields),
            fmt_=", ".join(fmt for _ in self._table_fields),
        )

    def _get_highscores_count_sql(self, *, game_mode: GameMode) -> str:
        """Get the SQL command to count the rows of the highscores table."""
        return f"SELECT COUNT(*) FROM {self.TABLES[game_mode]}"



# February 2020, Lewis Gaul

"""
Format bot messages.

"""

__all__ = (
    "format_highscores",
    "format_highscore_times",
    "format_kwargs",
    "format_matchups",
    "format_player_highscores",
    "format_player_info",
)

import datetime as dt
from typing import Iterable, List, Mapping, Optional, Tuple, Union

import pytz
import tabulate

from minegauler.app import highscores as hs
from minegauler.app.shared.types import Difficulty, GameMode

from .utils import Matchup, PlayerInfo


tabulate.PRESERVE_WHITESPACE = True


def format_highscores(highscores: Iterable[hs.HighscoreStruct]) -> str:
    return format_highscore_times([(h.name, h.elapsed) for h in highscores])


def format_highscore_times(highscores: Iterable[Tuple[str, float]]) -> str:
    lines = [
        f"{i+1:2d}. {h[0][:10]:<10s}  {h[1]:7.2f}" for i, h in enumerate(highscores)
    ]
    return "\n".join(lines)


def format_player_highscores(
    highscores: List[hs.HighscoreStruct], difficulty: Optional[Difficulty] = None
) -> List[str]:
    lines = []
    if highscores:
        last_played = format_timestamp(max(h.timestamp for h in highscores))
    else:
        last_played = "<n/a>"
    lines.append(f"Last game played on {last_played}")

    if not difficulty:
        for diff in [
            Difficulty.BEGINNER,
            Difficulty.INTERMEDIATE,
            Difficulty.EXPERT,
            Difficulty.MASTER,
            Difficulty.LUDICROUS,
        ]:
            hscores = [h.elapsed for h in highscores if h.difficulty is diff]
            if hscores:
                best = f"{min(hscores):.2f}"
            else:
                best = "None"
            line = "{}: {}".format(diff.name.capitalize(), best)
            lines.append(line)
    else:
        lines.append(f"Top {difficulty.name.lower()} times:")
        for h in highscores[:5]:
            line = "{:.2f} ({:.2f} 3bv/s) - {}".format(
                h.elapsed, h.bbbvps, format_timestamp(h.timestamp)
            )
            lines.append(line)

    return lines


def format_player_info(players: Iterable[PlayerInfo]) -> str:
    headers = [
        "Username",
        "Nickname",
        "Combined time",
        "Modes played",
        "Last highscore",
    ]
    types_available = len(GameMode) * (len(Difficulty) - 1) * 3 * 2
    data = [
        (
            p.username,
            p.nickname[:10],
            p.combined_time,
            f"{p.types_beaten:2d}/{types_available}",
            format_timestamp(p.last_highscore) if p.last_highscore else "None",
        )
        for p in sorted(players, key=lambda x: x.combined_time)
    ]
    return tabulate.tabulate(
        data,
        headers=headers,
        tablefmt="presto",
        stralign="center",
        numalign="center",
        floatfmt="7.2f",
    )


def format_kwargs(kwargs: Mapping) -> str:
    return ", ".join(f"{k}={v}" for k, v in kwargs.items())


def format_filters(
    *,
    game_mode: Optional[GameMode],
    difficulty: Optional[Union[str, Difficulty]],
    drag_select: Optional[bool],
    per_cell: Optional[int],
    no_difficulty=False,
) -> str:
    opts = {}
    if game_mode:
        opts["mode"] = game_mode.value
    if not no_difficulty:
        try:
            diff = difficulty.name.lower()
        except AttributeError:
            diff = difficulty if difficulty else "combined"
        opts["difficulty"] = diff
    if drag_select is not None:
        opts["drag-select"] = "on" if drag_select else "off"
    if per_cell is not None:
        opts["per-cell"] = per_cell
    return format_kwargs(opts)


def format_matchups(matchups: Iterable[Matchup]) -> List[str]:
    return [
        "{} ({:.2f}) vs {} ({:.2f}) - {:.2f}% difference".format(*m) for m in matchups
    ]


def format_timestamp(timestamp: float) -> str:
    tz = pytz.timezone("Europe/London")
    date = dt.datetime.fromtimestamp(timestamp).astimezone(tz)
    return date.strftime("%Y-%m-%d %H:%M")



# October 2021, Lewis Gaul

from typing import Iterable, List

from ...shared.types import CellContents, GameMode
from ..board import BoardBase
from .types import Coord


class Board(BoardBase):
    """A split-cell minesweeper board."""

    mode = GameMode.SPLIT_CELL

    def __init__(self, x_size: int, y_size: int):
        """
        :param x_size:
            Number of small cells in x direction.
        :param y_size:
            Number of small cells in y direction.
        """
        self.x_size = x_size
        self.y_size = y_size
        self._all_coords = {
            Coord(2 * x, 2 * y, False): CellContents.Unclicked
            for x in range(self.x_size // 2)
            for y in range(self.y_size // 2)
        }

    def __eq__(self, other) -> bool:
        if not isinstance(other, Board):
            return False
        if (self.x_size, self.y_size) != (other.x_size, other.y_size):
            return False
        return self.all_coords == other.all_coords

    def __getitem__(self, coord: Coord) -> CellContents:
        return self._all_coords[coord]

    def __setitem__(self, coord: Coord, obj: CellContents) -> None:
        # TODO: This seems desirable, but is used for displaying hit mines...
        # if coord not in self:
        #     raise KeyError(f"{coord} not in board")
        self._all_coords[coord] = obj

    def __contains__(self, coord: Coord):
        return coord in self._all_coords

    @property
    def all_coords(self) -> List[Coord]:
        return sorted(self._all_coords)

    @property
    def all_underlying_coords(self) -> List[Coord]:
        return [
            Coord(x, y, True) for x in range(self.x_size) for y in range(self.y_size)
        ]

    def copy(self) -> "Board":
        new_board = type(self)(self.x_size, self.y_size)
        new_board._all_coords = self._all_coords.copy()
        return new_board

    def get_nbrs(
        self, coord: Coord, *, include_origin: bool = False
    ) -> Iterable[Coord]:
        if coord not in self.all_coords:
            raise ValueError(f"{coord} not in board")
        x_min = max(0, coord.x - 1)
        y_min = max(0, coord.y - 1)
        if coord.is_split:
            x_max = min(self.x_size - 1, coord.x + 1)
            y_max = min(self.y_size - 1, coord.y + 1)
        else:
            x_max = min(self.x_size - 1, coord.x + 2)
            y_max = min(self.y_size - 1, coord.y + 2)
        nbrs = set()
        for i in range(x_min, x_max + 1):
            for j in range(y_min, y_max + 1):
                nbrs.add(self.get_coord_at(i, j))
        if not include_origin:
            nbrs.remove(coord)
        return sorted(nbrs)

    def get_coord_at(self, x: int, y: int) -> Coord:
        split = Coord(x, y, True)
        unsplit = Coord(x // 2 * 2, y // 2 * 2, False)
        if split in self._all_coords:
            return split
        elif unsplit in self._all_coords:
            return unsplit
        else:
            raise ValueError(f"Position out of bounds: ({x}, {y})")

    def split_coord(self, coord: Coord) -> None:
        self._all_coords.pop(coord)
        self._all_coords.update({c: CellContents.Unclicked for c in coord.split()})



# February 2022, Lewis Gaul

import runpy
import sys


if len(sys.argv) == 1:
    runpy.run_module("minegauler.app", run_name="__main__")
else:
    try:
        runpy.run_module("minegauler.cli", run_name="__main__")
    except ImportError:
        print("Expected no CLI arguments", file=sys.stderr)
        sys.exit(1)



# July 2018, Lewis Gaul

"""
Highscores window implementation.

Exports
-------
.. class:: HighscoresWindow
    Widget for displaying highscores.

"""

__all__ = ("HighscoresWindow",)

import logging
import time
from typing import Dict, List, Optional

from PyQt5.QtCore import (
    QAbstractTableModel,
    QModelIndex,
    QPoint,
    QRect,
    QSize,
    Qt,
    QVariant,
    pyqtSignal,
    pyqtSlot,
)
from PyQt5.QtGui import QCursor, QFont, QHideEvent
from PyQt5.QtWidgets import (
    QAbstractItemView,
    QAbstractScrollArea,
    QAction,
    QActionGroup,
    QDialog,
    QHBoxLayout,
    QLineEdit,
    QMenu,
    QTableView,
    QWidget,
    QWidgetAction,
)

from .. import highscores
from ..shared import utils
from . import state


logger = logging.getLogger(__name__)


class HighscoresWindow(QDialog):
    """A standalone highscores window."""

    def __init__(
        self,
        parent: Optional[QWidget],
        settings: highscores.HighscoreSettingsStruct,
        state_: state.HighscoreWindowState,
    ):
        super().__init__(parent)
        self._state = state_
        self.setWindowTitle("Highscores")
        self._model = HighscoresModel(self, self._state)
        self._table = HighscoresTable(self._model, self._state)
        self.setup_ui()
        self._model.sort_changed.connect(self._table.set_sort_indicator)
        self._table.add_filter.connect(self.set_filter)
        self._model.update_highscores_group(settings)

    def setup_ui(self) -> None:
        lyt = QHBoxLayout(self)
        # settings_frame = QFrame(self)
        # lyt.addWidget(settings_frame) #[currently not implemented]
        # settings_frame.setLineWidth(2)
        # settings_frame.setFrameShape(QFrame.StyledPanel)
        # Make highscores table.
        lyt.addWidget(self._table)
        # Make settings/filter panel.
        # lyt = QVBoxLayout(settings_frame)
        # settings_frame.setFixedSize(200, 200)

    def keyPressEvent(self, event):
        """Override the QWidget method for receiving key presses."""
        # Make enter/return/escape close the window.
        if event.key() in [Qt.Key_Return, Qt.Key_Enter, Qt.Key_Escape]:
            self.close()
        else:
            super().keyPressEvent(event)

    @pyqtSlot(str, str)
    def set_filter(self, filter_by: str, filter: str) -> None:
        if filter_by == "name":
            self._state.name_filter = filter
        elif filter_by == "flagging":
            self._state.flagging_filter = filter
        else:
            raise ValueError(f"Unrecognised header to filter by: '{filter_by}'")
        self._model.filter_and_sort()


class HighscoresModel(QAbstractTableModel):
    """Model handling sorting and filtering of a highscore group."""

    sort_changed = pyqtSignal(int)
    _HEADERS = ["name", "time", "3bv", "3bv/s", "date", "flagging"]

    def __init__(self, parent: Optional[QWidget], state_: state.HighscoreWindowState):
        super().__init__(parent)
        self._state: state.HighscoreWindowState = state_
        self._all_data: List[highscores.HighscoreStruct] = []
        self._displayed_data: List[highscores.HighscoreStruct] = []

    @property
    def _filters(self) -> Dict[str, Optional[str]]:
        return {
            "name": self._state.name_filter,
            "flagging": self._state.flagging_filter,
        }

    # -------------------------------------------------------------------------
    # Implement abstract methods
    # -------------------------------------------------------------------------
    def flags(self, index: QModelIndex) -> Qt.ItemFlags:
        return super().flags(index)

    def columnCount(self, parent: QModelIndex = QModelIndex()) -> int:
        if parent.isValid():
            return 0
        return len(self._HEADERS)

    def rowCount(self, parent: QModelIndex = QModelIndex()) -> int:
        if parent.isValid():
            return 0
        return len(self._displayed_data)

    def headerData(
        self, section: int, orientation: Qt.Orientation, role: Qt.ItemDataRole
    ) -> QVariant:
        bold_font = QFont("Sans-serif", 9)
        bold_font.setBold(True)
        # Horizontal header
        if orientation == Qt.Horizontal:
            header = self._HEADERS[section]
            if role == Qt.DisplayRole:
                return QVariant(header.capitalize())
            elif role == Qt.FontRole:
                if (
                    self._filters.get(header)
                    or self._state.sort_by == self._HEADERS[section]
                ):
                    return bold_font
            elif role == Qt.SizeHintRole:
                # Set size hint for 'Name' column width (minimum width).
                if header == "name":
                    return QSize(200, 0)
        # Vertical indexing
        elif orientation == Qt.Vertical:
            if role == Qt.DisplayRole:
                return QVariant(str(section + 1))
            elif role == Qt.FontRole:
                if section == self._get_active_row():
                    return bold_font
        return QVariant()

    def data(self, index: QModelIndex, role: Qt.ItemDataRole) -> QVariant:
        header = self._HEADERS[index.column()]
        if not index.isValid():
            return QVariant()
        elif role == Qt.DisplayRole:
            return QVariant(self._format_data(index.row(), header))
        elif role == Qt.TextAlignmentRole:
            return QVariant(Qt.AlignHCenter | Qt.AlignVCenter)
        elif role == Qt.FontRole:
            bold_font = QFont("Sans-serif", 9)
            bold_font.setBold(True)
            if index.row() == self._get_active_row():
                return bold_font
        return QVariant()

    def sort(self, column: int, order: Qt.SortOrder = Qt.DescendingOrder) -> None:
        header = self._HEADERS[column]
        if header not in ["time", "3bv/s"]:
            return
        self._state.sort_by = header
        self.filter_and_sort()
        self.sort_changed.emit(column)

    # -------------------------------------------------------------------------
    # Other methods
    # -------------------------------------------------------------------------
    def update_highscores_group(
        self, settings: highscores.HighscoreSettingsStruct
    ) -> None:
        """
        Change the data to be highscores for a different set of settings.
        """
        self._all_data = highscores.get_highscores(settings=settings)
        self.filter_and_sort()

    def _get_active_row(self) -> Optional[int]:
        """Get the index of the row containing the active highscore."""
        if self._state.current_highscore in self._displayed_data:
            return self._displayed_data.index(self._state.current_highscore)
        else:
            return None

    def _format_data(self, row: int, key: str) -> str:
        """Get the string to display in a given cell."""
        h = self._displayed_data[row]
        if key == "time":
            return f"{h.elapsed + 0.005 : 6.2f}"
        elif key == "3bv":
            return f"{h.bbbv:3d}"
        elif key == "3bv/s":
            return f"{h.bbbvps + 0.005 : .2f}"
        elif key == "name":
            return h.name
        elif key == "date":
            return time.strftime("%Y-%m-%d", time.localtime(h.timestamp))
        elif key == "flagging":
            return "F" if utils.is_flagging_threshold(h.flagging) else "NF"
        else:
            return "<unhandled>"

    def filter_and_sort(self):
        """Update the displayed data based on current filters/sorting."""
        self.layoutAboutToBeChanged.emit()
        self._displayed_data = highscores.filter_and_sort(
            self._all_data, self._state.sort_by, self._filters
        )
        # TODO: Should call changePersistentIndexList()?
        self.layoutChanged.emit()
        self.dataChanged.emit(QModelIndex(), QModelIndex())


class HighscoresTable(QTableView):
    """A table view for highscores."""

    add_filter = pyqtSignal(str, str)

    def __init__(self, model: HighscoresModel, state_: state.HighscoreWindowState):
        super().__init__()
        self._state: state.HighscoreWindowState = state_
        self.setModel(model)
        self._header = self.horizontalHeader()
        self._index = self.verticalHeader()
        self.setSizeAdjustPolicy(QAbstractScrollArea.AdjustToContents)
        # self.setMinimumWidth(500)
        self.setMinimumHeight(300)
        self.setStyleSheet(
            """
            background: rgb(215,215,255);
            font:       normal 9pt Sans-serif;
            """
        )
        self.setAlternatingRowColors(True)
        self.setSelectionMode(QAbstractItemView.NoSelection)
        self.setFocusPolicy(Qt.NoFocus)
        self.setCornerButtonEnabled(False)
        self.setSortingEnabled(True)
        # TODO: ResizeToContents is too slow.
        # self._header.setSectionResizeMode(QHeaderView.ResizeToContents)
        self._header.setSortIndicatorShown(True)
        self.set_sort_indicator()
        # Sort indicator is changed by clicking a header column, although in
        #  most cases this won't change the sorting, so the indicator needs to
        #  be changed back.
        self._header.sortIndicatorChanged.connect(self.set_sort_indicator)
        self._header.sectionClicked.connect(self.show_header_menu)
        # Set height of rows.
        # TODO: ResizeToContents is too slow.
        # self._index.setSectionResizeMode(QHeaderView.ResizeToContents)
        self._index.setSectionsClickable(False)
        self._filter_menu = QMenu(None)
        self._block_header_menu = False

    @property
    def _model(self) -> HighscoresModel:
        return self.model()

    def hideEvent(self, event: QHideEvent):
        self._filter_menu.close()

    def sizeHint(self) -> QSize:
        width = super().sizeHint().width()
        height = width / 1.61
        return QSize(width, height)

    def set_sort_indicator(self, *_):
        """Set the sort indicator to match the actual sorting."""
        self._header.setSortIndicator(
            self._model._HEADERS.index(self._state.sort_by), Qt.DescendingOrder
        )

    @pyqtSlot(int)
    def show_header_menu(self, col: int):
        """Pop up a mini entry menu for filtering on a column."""
        key = self._model._HEADERS[col]
        if key not in ["name", "flagging"] or self._block_header_menu:
            self._block_header_menu = False
            return
        self._filter_menu = QMenu(self.parent())
        self._filter_menu.aboutToHide.connect(self._check_mouse_on_header_menu_hide)

        def get_filter_cb(key, f):
            def cb():
                self.add_filter.emit(key, f)
                self._block_header_menu = False

            return cb

        if key == "flagging":
            group = QActionGroup(self)
            group.setExclusive(True)
            for filter_string in ["All", "F", "NF"]:
                action = QAction(filter_string, group, checkable=True)
                if filter_string == "All" and not self._state.flagging_filter:
                    action.setChecked(True)
                    filter_string = ""
                elif filter_string == self._state.flagging_filter:
                    action.setChecked(True)
                self._filter_menu.addAction(action)
                action.triggered.connect(get_filter_cb(key, filter_string))

        elif key == "name":
            # Make button for resetting filter (show all).
            all_action = QAction("All", checkable=True)
            if not self._model._filters["name"]:
                all_action.setChecked(True)
            all_action.triggered.connect(get_filter_cb(key, None))
            self._filter_menu.addAction(all_action)  # add to menu
            self._filter_menu.addSeparator()  # patch highlighting with mouse movement
            ## Make entry bar for name filter.
            # Create the entry bar with the existing filter as the text.
            if self._state.name_hint:
                # Name in entry bar, if any.
                text = self._state.name_hint
            else:
                # If no name in entry bar, name currently filtered by.
                text = self._state.name_filter
            entry = QLineEdit(text, self)
            entry.selectAll()  # select all the text
            # Set focus to the entry bar when the menu is opened.
            self._filter_menu.aboutToShow.connect(entry.setFocus)

            def set_name_filter():
                self.add_filter.emit(key, entry.text().strip())
                self._filter_menu.hide()
                self._block_header_menu = False

            entry.returnPressed.connect(set_name_filter)  # enter applies filter
            name_action = QWidgetAction(self._filter_menu)  # to contain QLineEdit
            name_action.setDefaultWidget(entry)  # set widget on QWidgetAction
            self._filter_menu.addAction(name_action)  # add to menu
        self._filter_menu.index = col
        # Display menu in appropriate position, below header in column 'col'
        headerPos = self.mapToGlobal(self._header.pos())
        posY = headerPos.y() + self._header.height()
        posX = headerPos.x() + self._header.sectionPosition(col)
        pos = QPoint(posX, posY)
        self._filter_menu.exec_(pos)  # modal dialog
        self.resizeRowsToContents()

    def _check_mouse_on_header_menu_hide(self):
        header_pos = self.mapToGlobal(self._header.pos())
        x_min = header_pos.x() + sum(
            [self._header.sectionSize(i) for i in range(self._filter_menu.index)]
        )
        width = self._header.sectionSize(self._filter_menu.index)
        y_min = header_pos.y()
        height = self.mapToGlobal(self._index.pos()).y() - y_min
        if QRect(x_min, y_min, width, height).contains(QCursor.pos()):
            self._block_header_menu = True



# March 2018, Lewis Gaul

__all__ = (
    "GameBase",
    "GameNotStartedError",
    "check_game_started",
)

import abc
import functools
import logging
import math
import time
from typing import Callable, Dict, Iterable, List, Mapping, Optional, Tuple, Type, Union

from ..shared.types import CellContents, Coord, Difficulty, GameMode, GameState
from .board import BoardBase
from .minefield import MinefieldBase


logger = logging.getLogger(__name__)


def _check_coord(method: Callable) -> Callable:
    """
    Wrap a method that takes a coord to check it is inside the valid range.

    :raise ValueError:
        If the coord is not valid.
    """

    @functools.wraps(method)
    def wrapped(self: "GameBase", coord: Coord, *args, **kwargs):
        if not 0 <= coord.x < self.x_size or not 0 <= coord.y < self.y_size:
            raise ValueError(
                f"Coordinate is out of bounds, should be between (0,0) and "
                f"({self.x_size-1}, {self.y_size-1})"
            )
        return method(self, coord, *args, **kwargs)

    return wrapped


def _ignore_decorator_helper(conditions_sense, game_state, cell_state) -> Callable:
    # If the cell state is specified it is assumed the coord is passed in
    # as the first arg.
    def decorator(method: Callable) -> Callable:
        if game_state is None:
            game_states = []
        elif type(game_state) is GameState:
            game_states = [game_state]
        else:
            game_states = game_state

        if cell_state is None:
            cell_states = []
        elif cell_state in CellContents.items:
            cell_states = [cell_state]
        else:
            cell_states = cell_state

        @functools.wraps(method)
        def wrapped(game: "GameBase", coord: Coord = None, *args, **kwargs):
            conditions = [any(game.state is s for s in game_states)]
            if cell_states:
                conditions.append(
                    any(game.board[coord].is_type(s) for s in cell_states)
                )

            # fmt: off
            if (
                any(conditions) and conditions_sense
                or not all(conditions) and not conditions_sense
            ):
                return
            # fmt: on

            return method(game, coord, *args, **kwargs)

        return wrapped

    return decorator


def _ignore_if(
    *,
    game_state: Optional[Union[GameState, Iterable[GameState]]] = None,
    cell_state: Optional[Union[CellContents, Iterable[CellContents]]] = None,
) -> Callable:
    """
    Return a decorator which prevents a method from running if any of the given
    parameters are satisfied.

    Arguments:
    game_state=None (GameState | (GameState, ...) | None)
        A game state or iterable of game states to match against.
    cell_state=None (subclass of CellContents | (..., ...) | None)
        A cell contents type or iterable of the same to match against. The
        decorated method must take the cell coordinate as the first argument.
    """
    return _ignore_decorator_helper(True, game_state, cell_state)


def _ignore_if_not(
    *,
    game_state: Optional[Union[GameState, Iterable[GameState]]] = None,
    cell_state: Optional[Union[CellContents, Iterable[CellContents]]] = None,
) -> Callable:
    """
    Return a decorator which prevents a method from running if any of the given
    parameters are not satisfied.

    Arguments:
    game_state=None (GameState | (GameState, ...) | None)
        A game state or iterable of game states to match against.
    cell_state=None (subclass of CellContents | (..., ...) | None)
        A cell contents type or iterable of the same to match against. The
        decorated method must take the cell coordinate as the first argument.
    """
    return _ignore_decorator_helper(False, game_state, cell_state)


def check_game_started(method: Callable) -> Callable:
    """Check the game has been started, raising an error if not."""

    @functools.wraps(method)
    def wrapped(self: "GameBase", *args, **kwargs):
        if self.state is GameState.READY:
            raise GameNotStartedError("Minefield may not yet be created")
        assert self.start_time is not None
        return method(self, *args, **kwargs)

    return wrapped


class GameNotStartedError(Exception):
    """Game has not been started, so no minefield has been created."""


class GameBase(metaclass=abc.ABCMeta):
    """Representation of a minesweeper game, generic on the game mode."""

    mode: GameMode
    minefield_cls: Type[MinefieldBase]
    board_cls: Type[BoardBase]

    _diff_pairs: List[Tuple[Difficulty, Tuple[int, int, int]]]

    def __init__(
        self,
        *,
        x_size: int,
        y_size: int,
        mines: int,
        per_cell: int = 1,
        lives: int = 1,
        first_success: bool = False,
    ):
        self.x_size: int = x_size
        self.y_size: int = y_size
        self.board: BoardBase = self._make_board()
        self.mf: MinefieldBase = self.minefield_cls(
            self.board.all_underlying_coords, mines=mines, per_cell=per_cell
        )
        self.minefield_known: bool = False
        self.lives: int = lives
        self.first_success: bool = first_success
        self.start_time: Optional[float] = None
        self.end_time: Optional[float] = None
        self.state: GameState = GameState.READY
        self.mines_remaining: int = self.mines
        self.lives_remaining: int = self.lives
        self._num_flags: int = 0

        self._cell_updates: Dict[Coord, CellContents] = {}

    @abc.abstractmethod
    def _make_board(self) -> BoardBase:
        raise NotImplementedError

    @abc.abstractmethod
    def get_rem_3bv(self) -> int:
        """Calculate the minimum remaining number of clicks needed to solve."""
        raise NotImplementedError

    @abc.abstractmethod
    def _populate_minefield(self, coord: Coord) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def _is_complete(self) -> bool:
        raise NotImplementedError

    @abc.abstractmethod
    def _select_cell_action(self, coord: Coord) -> None:
        raise NotImplementedError

    @classmethod
    def from_minefield(cls, mf: MinefieldBase, **kwargs) -> "GameBase":
        self = cls(
            x_size=mf.x_size,
            y_size=mf.y_size,
            mines=mf.mines,
            per_cell=mf.per_cell,
            **kwargs,
        )
        self.mf = mf
        if mf.populated:
            self.minefield_known = True
        return self

    @classmethod
    def difficulty_to_values(cls, diff: Difficulty) -> Tuple[int, int, int]:
        try:
            return dict(cls._diff_pairs)[diff]
        except KeyError:
            raise ValueError(f"Unknown difficulty: {diff}") from None

    @classmethod
    def difficulty_from_values(cls, x_size: int, y_size: int, mines: int) -> Difficulty:
        mapping = dict((x[1], x[0]) for x in cls._diff_pairs)
        try:
            return mapping[(x_size, y_size, mines)]
        except KeyError:
            return Difficulty.CUSTOM

    @property
    def mines(self) -> int:
        return self.mf.mines

    @property
    def difficulty(self) -> Difficulty:
        return self.difficulty_from_values(self.x_size, self.y_size, self.mines)

    @property
    def per_cell(self) -> int:
        return self.mf.per_cell

    def get_prop_complete(self) -> float:
        """Calculate the progress of solving the board using 3bv."""
        rem_3bv = self.get_rem_3bv()
        try:
            return (self.mf.bbbv - rem_3bv) / self.mf.bbbv
        except ZeroDivisionError:
            # This can only occur for created boards with no safe cells,
            # which can technically never be completed.
            return 0

    def get_3bvps(self) -> float:
        """Calculate the 3bv/s based on current progress."""
        if self.get_elapsed() == 0:
            return math.inf
        return self.mf.bbbv * self.get_prop_complete() / self.get_elapsed()

    @check_game_started
    def get_elapsed(self) -> float:
        """Get the elapsed game time."""
        if self.end_time:
            return self.end_time - self.start_time
        else:
            return time.time() - self.start_time

    def get_flag_proportion(self) -> float:
        """Get the proportion of the mines that have been flagged."""
        if self.mines == 0:
            if self._num_flags > 0:
                return math.inf
            else:
                return 0
        return self._num_flags / self.mines

    # ---------------------
    # Other methods
    # ---------------------
    def _set_cell(self, coord: Coord, state: CellContents) -> None:
        """
        Set the contents of a cell and store the update.

        :param coord:
            The coordinate of the cell to set.
        :param state:
            The state to set the cell to.
        """
        self.board[coord] = state
        self._cell_updates[coord] = state

    def _check_for_completion(self) -> None:
        """
        Check if game is complete by comparing the board to the minefield's
        completed board. If it is, display flags in remaining unclicked cells.
        """
        if self._is_complete():
            logger.info("Game won")

            self.end_time = time.time()
            self.state = GameState.WON
            self.mines_remaining = 0

            for c in self.mf.mine_coords:
                self._set_cell(c, CellContents.Flag(self.mf[c]))

    @_check_coord
    @_ignore_if_not(
        game_state=(GameState.READY, GameState.ACTIVE),
        cell_state=CellContents.Unclicked,
    )
    def select_cell(self, coord: Coord) -> Mapping[Coord, CellContents]:
        """Perform the action of selecting/clicking a cell."""
        just_started = False
        if self.state is GameState.READY:
            if not self.mf.populated:
                self._populate_minefield(coord)
            self.state = GameState.ACTIVE
            self.start_time = time.time()
            just_started = True
        self._select_cell_action(coord)
        if not self.state.finished():
            self._check_for_completion()
        if self.state.finished() and just_started:
            self.end_time = self.start_time
        try:
            return self._cell_updates
        finally:
            self._cell_updates = {}

    @_check_coord
    @_ignore_if_not(
        game_state=(GameState.READY, GameState.ACTIVE),
        cell_state=(CellContents.Flag, CellContents.Unclicked),
    )
    def set_cell_flags(
        self, coord: Coord, nr_flags: int
    ) -> Mapping[Coord, CellContents]:
        """Set the number of flags in a cell."""
        if nr_flags < 0 or nr_flags > self.per_cell:
            raise ValueError(
                f"Invalid number of flags ({nr_flags}) - should be between 0 and "
                f"{self.per_cell}"
            )

        old_nr_flags = (
            0 if self.board[coord] is CellContents.Unclicked else self.board[coord].num
        )
        if nr_flags == 0:
            self._set_cell(coord, CellContents.Unclicked)
        else:
            self._set_cell(coord, CellContents.Flag(nr_flags))
        self.mines_remaining += old_nr_flags - nr_flags
        self._num_flags += nr_flags - old_nr_flags

        try:
            return self._cell_updates
        finally:
            self._cell_updates = {}

    @_check_coord
    @_ignore_if_not(game_state=GameState.ACTIVE, cell_state=CellContents.Num)
    def chord_on_cell(self, coord: Coord) -> Mapping[Coord, CellContents]:
        """Chord on a cell that contains a revealed number."""
        nbrs = self.board.get_nbrs(coord)
        num_flagged_nbrs = sum(
            [self.board[c].num for c in nbrs if self.board[c].is_mine_type()]
        )
        logger.debug(
            "%s flagged mine(s) around clicked cell showing number %s",
            num_flagged_nbrs,
            self.board[coord],
        )

        unclicked_nbrs = [c for c in nbrs if self.board[c] is CellContents.Unclicked]
        if (
            self.board[coord] != CellContents.Num(num_flagged_nbrs)
            or not unclicked_nbrs
        ):
            return {}

        logger.info("Successful chording, selecting cells %s", unclicked_nbrs)
        for c in unclicked_nbrs:
            self._select_cell_action(c)

        if self.state != GameState.LOST:
            self._check_for_completion()

        try:
            return self._cell_updates
        finally:
            self._cell_updates = {}



# December 2018, Lewis Gaul

"""
Available imports from the package.

"""

__all__ = ("MinegaulerGUI", "init_app", "run_app", "state")

import signal
import sys

from PyQt5.QtCore import QTimer, pyqtRemoveInputHook
from PyQt5.QtWidgets import QApplication, QWidget

from . import state
from .main_window import MinegaulerGUI


_app = None
_timer = None


def init_app() -> None:
    """Initialise the GUI application. This function is idempotent."""
    global _app
    if not QApplication.instance():
        _app = QApplication(sys.argv)
    else:
        _app = QApplication.instance()
    pyqtRemoveInputHook()


def run_app(gui: QWidget) -> int:
    """
    Run the GUI application.

    :param gui:
        The main GUI widget.
    :return:
        Exit code.
    """
    global _timer

    if not _app:
        raise RuntimeError("App must be initialised before being run")

    gui.show()

    # Create a timer to run Python code periodically, so that Ctrl+C can be caught.
    signal.signal(signal.SIGINT, lambda *args: gui.close())
    _timer = QTimer()
    _timer.timeout.connect(lambda: None)
    _timer.start(100)

    return _app.exec_()



# April 2018, Lewis Gaul

"""
Base GUI application implementation.

Exports
-------
.. class:: MinegaulerGUI
    Main window widget.

"""

__all__ = ("MinegaulerGUI",)

import functools
import logging
import pathlib
import textwrap
import traceback
from typing import Callable, Dict, Mapping, Optional

from PyQt5.QtCore import QSize, Qt, pyqtSignal
from PyQt5.QtGui import QFocusEvent, QFont, QIcon, QKeyEvent
from PyQt5.QtWidgets import (
    QWIDGETSIZE_MAX,
    QAction,
    QActionGroup,
    QApplication,
    QDialog,
    QFileDialog,
    QFrame,
    QGridLayout,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QMainWindow,
    QMenu,
    QMenuBar,
    QMessageBox,
    QPushButton,
    QSizePolicy,
    QSlider,
    QSpinBox,
    QVBoxLayout,
    QWidget,
)

from .. import api, highscores, paths
from ..highscores import HighscoreSettingsStruct, HighscoreStruct
from ..shared.types import (
    CellContents,
    CellImageType,
    Coord,
    Difficulty,
    GameMode,
    GameState,
    PathLike,
    UIMode,
)
from ..shared.utils import GUIOptsStruct, format_timestamp
from . import highscores as highscores_ui
from . import minefield, panel, state, utils
from .minefield import simulate


logger = logging.getLogger(__name__)


def _msg_popup(
    parent: QWidget, icon: QMessageBox.Icon, title: str, msg: Optional[str] = None
) -> None:
    """Open a modal popup with a simple message."""
    logger.debug(
        "Opening popup with message: %s", msg if len(msg) <= 50 else msg[:47] + "..."
    )
    popup = QMessageBox(parent)
    popup.setIcon(icon)
    popup.setWindowTitle(title)
    if msg:
        popup.setText(msg)
    popup.exec_()


class _BaseMainWindow(QMainWindow):
    """
    Base class for the application implementing the general layout.
    """

    BODY_FRAME_WIDTH = 10

    def __init__(self, parent: Optional[QWidget], title: Optional[str] = None):
        """
        :param title:
            The window title.
        """
        super().__init__(parent)
        self._menubar: QMenuBar = self.menuBar()
        self._game_menu: QMenu = self._menubar.addMenu("Game")
        self._opts_menu: QMenu = self._menubar.addMenu("Options")
        self._help_menu: QMenu = self._menubar.addMenu("Help")
        self._panel_frame: QFrame
        self._body_frame: QFrame
        self._footer_frame: QFrame
        self._panel_widget: Optional[QWidget]
        self._body_widget: Optional[QWidget]
        self._footer_widget: Optional[QWidget]
        self._icon: QIcon = QIcon(str(paths.IMG_DIR / "icon.ico"))
        self.setWindowTitle(title)
        self.setWindowIcon(self._icon)
        # Disable maximise button
        self.setWindowFlags(self.windowFlags() | Qt.CustomizeWindowHint)
        self.setWindowFlags(self.windowFlags() & ~Qt.WindowMaximizeButtonHint)
        self._setup_ui()
        # Keep track of all non-modal subwindows that are open.
        self._open_subwindows: Dict[str, QWidget] = {}

    # --------------------------------------------------------------------------
    # UI setup
    # --------------------------------------------------------------------------
    def _setup_ui(self):
        """
        Set up the layout of the main window GUI.
        """
        # QMainWindow objects have a central widget to be set.
        central_widget = QWidget(self)
        central_widget.setSizePolicy(QSizePolicy.Ignored, QSizePolicy.Ignored)
        self.setCentralWidget(central_widget)
        vlayout = QVBoxLayout(central_widget)
        vlayout.setContentsMargins(0, 0, 0, 0)
        vlayout.setSpacing(0)
        # Top panel widget.
        self._panel_frame = QFrame(central_widget)
        self._panel_frame.setFrameShadow(QFrame.Sunken)
        self._panel_frame.setFrameShape(QFrame.Panel)
        self._panel_frame.setLineWidth(2)
        self._panel_frame.setSizePolicy(QSizePolicy.Ignored, QSizePolicy.Fixed)
        vlayout.addWidget(self._panel_frame)
        # Main body widget config - use horizontal layout for centre alignment.
        hstretch = QHBoxLayout()
        hstretch.addStretch()  # left-padding for centering
        self._body_frame = QFrame(central_widget)
        self._body_frame.setFrameShadow(QFrame.Raised)
        self._body_frame.setFrameShape(QFrame.Box)
        self._body_frame.setLineWidth(self.BODY_FRAME_WIDTH // 2)
        hstretch.addWidget(self._body_frame)
        hstretch.addStretch()  # right-padding for centering
        vlayout.addLayout(hstretch)
        # Name entry bar underneath the minefield
        self._footer_frame = QFrame(central_widget)
        vlayout.addWidget(self._footer_frame)

    def set_panel_widget(self, widget: QWidget) -> None:
        """
        Set the widget to occupy the top panel.

        Arguments:
        widget (QWidget)
            The widget instance to place in the top panel.
        """
        lyt = QVBoxLayout(self._panel_frame)
        lyt.setContentsMargins(0, 0, 0, 0)
        lyt.addWidget(widget)
        self._panel_widget = widget

    def set_body_widget(self, widget: QWidget) -> None:
        """
        Set the widget to occupy the main section of the GUI.

        Arguments:
        widget (QWidget)
            The widget instance to place in the body.
        """
        lyt = QVBoxLayout(self._body_frame)
        lyt.setContentsMargins(0, 0, 0, 0)
        lyt.addWidget(widget)
        self._body_widget = widget

    def set_footer_widget(self, widget: QWidget) -> None:
        """
        Set the widget to occupy the lower bar of the GUI.

        Arguments:
        widget (QWidget)
            The widget instance to place in the lower bar.
        """
        lyt = QVBoxLayout(self._footer_frame)
        lyt.setContentsMargins(0, 0, 0, 0)
        lyt.addWidget(widget)
        self._footer_widget = widget


class _MinegaulerGUIMeta(type(api.AbstractListener), type(_BaseMainWindow)):
    """Combined metaclass for the MinegaulerGUI class."""


class MinegaulerGUI(
    api.AbstractListener, _BaseMainWindow, metaclass=_MinegaulerGUIMeta
):
    """The main Minegauler GUI window."""

    def __init__(self, ctrlr: api.AbstractController, initial_state: state.State):
        """
        :param ctrlr:
            The core controller.
        :param initial_state:
            Initial application state.
        """
        super().__init__(None, "Minegauler")
        self._ctrlr: api.AbstractController = ctrlr
        self._state: state.State = initial_state.deepcopy()

        self._create_menu_action: QAction
        self._diff_menu_actions: Dict[Difficulty, QAction] = dict()
        self._populate_menubars()
        self._menubar.setFixedHeight(self._menubar.sizeHint().height())
        self._panel_widget = panel.PanelWidget(self, self._state)
        self._mf_widget = minefield.MinefieldWidget(self, self._ctrlr, self._state)
        self.set_panel_widget(self._panel_widget)
        self.set_body_widget(self._mf_widget)
        self._name_entry_widget = _NameEntryBar(self, self._state.name)
        self.set_footer_widget(self._name_entry_widget)
        self._update_size()

        self._panel_widget.clicked.connect(self._ctrlr.new_game)
        self._mf_widget.at_risk_signal.connect(self._panel_widget.at_risk)
        self._mf_widget.no_risk_signal.connect(self._panel_widget.no_risk)
        self._mf_widget.size_changed.connect(self._update_size)
        self._name_entry_widget.name_updated_signal.connect(self._set_name)

    # ----------------------------------
    # Implemented abstractmethods
    # ----------------------------------
    def reset(self) -> None:
        """
        Called to indicate the GUI state should be soft-reset.

        This is distinct from a factory reset (settings are not changed).
        """
        self._state.game_status = GameState.READY
        self._panel_widget.reset()
        self._mf_widget.reset()

    def resize_minefield(self, x_size: int, y_size: int) -> None:
        """
        Called to indicate the board shape has changed.

        :param x_size:
            The number of columns.
        :param y_size:
            The number of rows.
        """
        # Game state needs updating first for state changes to be applied.
        if self._state.game_status is not GameState.READY:
            self.reset()
        self._state.x_size = x_size
        self._state.y_size = y_size
        self._mf_widget.reshape(x_size, y_size)

    def set_mines(self, mines: int) -> None:
        """
        Called to indicate the base number of mines has changed.

        :param mines:
            The number of mines.
        """
        self._state.mines = mines

    def set_difficulty(self, diff: Difficulty) -> None:
        """Called to indicate the difficulty has changed."""
        self._state.difficulty = diff
        self._diff_menu_actions[diff].setChecked(True)

    def update_cells(self, cell_updates: Mapping[Coord, CellContents]) -> None:
        """
        Called to indicate some cells have changed state.

        :param cell_updates:
            A mapping of cell coordinates to their new state.
        """
        self._mf_widget.update_cells(cell_updates)

    def update_game_state(self, game_state: GameState) -> None:
        """Called to indicate the game state has changed."""
        self._state.game_status = game_state
        self._state.highscores_state.current_highscore = None
        self._panel_widget.update_game_state(game_state)
        if game_state.finished():
            self._mf_widget._raise_all_sunken_cells()
            self._handle_finished_game()

    def update_mines_remaining(self, mines_remaining: int) -> None:
        """
        Called to indicate the number of remaining mines has changed.

        :param mines_remaining:
            The remaining number of mines.
        """
        self._panel_widget.set_mines_counter(mines_remaining)

    def ui_mode_changed(self, mode: UIMode) -> None:
        """
        Called to indicate the UI mode has changed.

        :param mode:
            The mode to change to.
        """
        super().ui_mode_changed(mode)
        self._create_menu_action.setChecked(mode is UIMode.CREATE)

    def game_mode_about_to_change(self, mode: GameMode) -> None:
        """Called to indicate the game mode is about to change."""
        super().game_mode_about_to_change(mode)
        self._mf_widget.switch_mode(mode)

    def game_mode_changed(self, mode: GameMode) -> None:
        """Called to indicate the game mode has just changed."""
        super().game_mode_changed(mode)
        self._state.game_mode = mode
        # TODO: Update the game mode radiobutton
        self.reset()

    def handle_exception(self, method: str, exc: Exception) -> None:
        logger.error(
            "Error occurred when calling %s() from backend:\n%s\n%s",
            method,
            "".join(traceback.format_exception(None, exc, exc.__traceback__)),
            exc,
        )
        raise RuntimeError(exc).with_traceback(exc.__traceback__)

    # --------------------------------------------------------------------------
    # Qt method overrides
    # --------------------------------------------------------------------------
    def sizeHint(self) -> QSize:
        width = self._body_frame.sizeHint().width()
        height = (
            self._menubar.sizeHint().height()
            + self._panel_frame.sizeHint().height()
            + self._body_frame.sizeHint().height()
            + self._name_entry_widget.sizeHint().height()
        )
        return QSize(width, height)

    def minimumSizeHint(self) -> QSize:
        width = self._panel_frame.minimumSizeHint().width()
        height = (
            self._menubar.sizeHint().height()
            + self._panel_frame.sizeHint().height()
            + 100
            + self._name_entry_widget.sizeHint().height()
        )
        return QSize(width, height)

    def mousePressEvent(self, event):
        super().mousePressEvent(event)
        self._name_entry_widget.clearFocus()

    # --------------------------------------------------------------------------
    # Other methods
    # --------------------------------------------------------------------------
    def _update_size(self):
        """Update the window size."""
        self.setMinimumSize(self.minimumSizeHint())
        self.setMaximumSize(self.sizeHint())
        self.resize(self.sizeHint())
        self.adjustSize()
        # print(self.sizeHint(), self.size())
        # print(self._menubar.sizeHint(), self._menubar.size())
        # print(self._panel_frame.sizeHint(), self._panel_frame.size())
        # print(self._body_frame.sizeHint(), self._body_frame.size())
        # print(self._mf_widget.sizeHint(), self._mf_widget.size())
        # print()

    def _populate_menubars(self) -> None:
        """Fill in the menubars."""
        # ----------
        # Game menu
        # ----------
        # New game (F2)
        new_game_act = self._game_menu.addAction("New game", self._ctrlr.new_game)
        new_game_act.setShortcut("F2")

        # Replay game (F3)
        replay_act = self._game_menu.addAction("Replay", self._ctrlr.restart_game)
        replay_act.setShortcut("F3")

        # Create board
        def switch_create_mode(checked: bool):
            mode = UIMode.CREATE if checked else UIMode.GAME
            self._state.ui_mode = mode
            self._ctrlr.switch_ui_mode(mode)

        self._create_menu_action = create_act = QAction(
            "Create board",
            self,
            checkable=True,
            checked=self._state.ui_mode is UIMode.CREATE,
        )
        self._game_menu.addAction(create_act)
        create_act.triggered.connect(switch_create_mode)

        # Save board
        self._game_menu.addAction("Save board", self._open_save_board_modal)

        # Load board
        self._game_menu.addAction("Load board", self._open_load_board_modal)

        self._game_menu.addSeparator()

        # Current info (F4)
        info_act = self._game_menu.addAction(
            "Current game info", self._open_current_info_modal
        )
        info_act.setShortcut("F4")

        # Solver
        # - Probabilities (F5)
        # - Auto flag (Ctrl+F)
        # - Auto click (Ctrl+Enter)

        # Highscores (F6)
        highscores_act = self._game_menu.addAction(
            "Highscores", self.open_highscores_window
        )
        highscores_act.setShortcut("F6")

        # Stats (F7)

        self._game_menu.addSeparator()

        # Difficulty radiobuttons
        # - Beginner (b)
        # - Intermediate (i)
        # - Expert (e)
        # - Master (m)
        # - Custom (c)
        diff_group = QActionGroup(self)
        diff_group.setExclusive(True)
        for diff in Difficulty:
            diff_act = QAction(diff.name.capitalize(), diff_group, checkable=True)
            self._diff_menu_actions[diff] = diff_act
            self._game_menu.addAction(diff_act)
            diff_act.id = diff
            if diff is self._state.difficulty:
                diff_act.setChecked(True)
            diff_act.triggered.connect(
                lambda _: self._change_difficulty(diff_group.checkedAction().id)
            )
            diff_act.setShortcut(diff.value)

        self._game_menu.addSeparator()

        # Play highscore
        # self._game_menu.addAction("Play highscore", self._open_play_highscore_modal)

        # Zoom
        self._game_menu.addAction("Button size", self._open_zoom_modal)

        # Themes
        themes = {
            "Standard": {
                CellImageType.BUTTONS: "Standard",
                CellImageType.MARKERS: "Standard",
                CellImageType.NUMBERS: "Standard",
            },
            "Butterfly": {
                CellImageType.BUTTONS: "Butterfly",
                CellImageType.MARKERS: "Standard",
                CellImageType.NUMBERS: "Standard",
            },
            "Halloween": {
                CellImageType.BUTTONS: "Dark",
                CellImageType.MARKERS: "Halloween",
                CellImageType.NUMBERS: "Dark",
            },
            "Textured": {
                CellImageType.BUTTONS: "Textured",
                CellImageType.MARKERS: "Standard",
                CellImageType.NUMBERS: "Standard",
            },
            "Christmas": {
                CellImageType.BUTTONS: "Christmas",
                CellImageType.MARKERS: "Christmas",
                CellImageType.NUMBERS: "Dark",
            },
        }

        themes_menu = QMenu("Themes", self)
        self._game_menu.addMenu(themes_menu)
        group = QActionGroup(self)
        group.setExclusive(True)
        for theme_name, theme_styles in themes.items():
            theme_act = QAction(theme_name, self, checkable=True)
            themes_menu.addAction(theme_act)
            group.addAction(theme_act)
            if theme_styles == self._state.styles:
                theme_act.setChecked(True)
            theme_act.triggered.connect(
                functools.partial(self._change_styles, theme_styles)
            )

        # Advanced options
        self._game_menu.addAction("Advanced options", self._open_advanced_opts_modal)

        self._game_menu.addSeparator()

        # Factory reset
        # TODO : Factory reset should also reset files such as highscores, settings
        # boards, so this menu button is disabled until that is implemented.
        # self._game_menu.addAction("Factory reset", self.factory_reset)

        # Exit (Alt+F4)
        self._game_menu.addAction("Exit", self.close, shortcut="Alt+F4")

        # ----------
        # Options menu
        # ----------
        # First-click success
        def toggle_first_success():
            new_val = not self._state.pending_first_success
            self._state.first_success = new_val
            self._ctrlr.set_first_success(new_val)

        first_act = QAction(
            "Safe start", self, checkable=True, checked=self._state.first_success
        )
        self._opts_menu.addAction(first_act)
        first_act.triggered.connect(toggle_first_success)

        # Drag select
        drag_act = self._opts_menu.addAction(
            "Drag select",
            lambda: setattr(
                self._state, "drag_select", not self._state.pending_drag_select
            ),
        )
        drag_act.setCheckable(True)
        drag_act.setChecked(self._state.drag_select)

        # Max mines per cell option
        def get_change_per_cell_func(n):
            def change_per_cell():
                self._state.per_cell = n
                self._ctrlr.set_per_cell(n)

            return change_per_cell

        per_cell_menu = self._opts_menu.addMenu("Max per cell")
        per_cell_group = QActionGroup(self)
        per_cell_group.setExclusive(True)
        for i in range(1, 4):
            action = QAction(str(i), self, checkable=True)
            per_cell_menu.addAction(action)
            per_cell_group.addAction(action)
            if self._state.per_cell == i:
                action.setChecked(True)
            action.triggered.connect(get_change_per_cell_func(i))

        self._opts_menu.addSeparator()

        game_mode_menu = self._opts_menu.addMenu("Game mode")
        game_mode_group = QActionGroup(self)
        game_mode_group.setExclusive(True)
        for mode in GameMode:
            action = QAction(
                mode.name.capitalize().replace("_", " "), self, checkable=True
            )
            game_mode_menu.addAction(action)
            game_mode_group.addAction(action)
            if self._state.game_mode is mode:
                action.setChecked(True)
            action.triggered.connect(functools.partial(self._change_game_mode, mode))

        # ----------
        # Help menu
        # ----------
        for file, shortcut in [
            ("rules", None),
            ("features", None),
            ("tips", None),
            ("about", "F1"),
        ]:
            act = QAction(file.capitalize(), self)
            self._help_menu.addAction(act)
            act.triggered.connect(
                functools.partial(
                    self._open_text_popup,
                    file.capitalize(),
                    paths.FILES_DIR / f"{file}.txt",
                )
            )
            if shortcut:
                act.setShortcut(shortcut)

        self._help_menu.addSeparator()

        retrieve_act = QAction("Retrieve highscores", self)
        retrieve_act.triggered.connect(self._open_retrieve_highscores_modal)
        self._help_menu.addAction(retrieve_act)

    def _change_difficulty(self, diff: Difficulty) -> None:
        """
        Change the difficulty via the core controller.

        :param diff:
            The difficulty.
        """
        if diff is Difficulty.CUSTOM:
            # Undo changing the menu radiobutton because the board hasn't been
            # changed yet.
            self._diff_menu_actions[self._state.difficulty].setChecked(True)
            logger.info("Opening popup window for selecting custom board")
            self._open_custom_board_modal()
            return
        else:
            self._ctrlr.set_difficulty(diff)

    def _change_styles(self, styles: Mapping[CellImageType, str]) -> None:
        for grp in styles:
            self._state.styles[grp] = styles[grp]
            self._mf_widget.update_style(grp, styles[grp])

    def _change_game_mode(self, mode: GameMode) -> None:
        self._ctrlr.switch_game_mode(mode)

    def _set_name(self, name: str) -> None:
        self._state.name = name
        self._state.highscores_state.name_hint = name

    def _handle_finished_game(self) -> None:
        """Called once when a game ends."""
        # TODO: Manually processing events here as getting game info can be slow...
        QApplication.processEvents()
        info: api.GameInfo = self._ctrlr.get_game_info()
        assert info.game_state.finished()
        assert info.started_info is not None

        self._panel_widget.timer.set_time(int(info.started_info.elapsed + 1))

        # Store the highscore if the game was won.
        if (
            info.game_state is GameState.WON
            and info.difficulty is not Difficulty.CUSTOM
            and not info.minefield_known
        ):
            assert info.started_info.prop_complete == 1
            highscore = HighscoreStruct(
                game_mode=info.mode,
                difficulty=info.difficulty,
                per_cell=info.per_cell,
                timestamp=int(info.started_info.start_time),
                elapsed=info.started_info.elapsed,
                bbbv=info.started_info.bbbv,
                bbbvps=info.started_info.bbbvps,
                drag_select=self._state.drag_select,
                name=self._state.name,
                flagging=info.started_info.prop_flagging,
            )
            try:
                highscores.insert_highscore(highscore)
            except Exception:
                logger.exception("Error inserting highscore")
            self._state.highscores_state.current_highscore = highscore
            # Check whether to pop up the highscores window.
            # TODO: This is too slow...
            try:
                new_best = highscores.is_highscore_new_best(
                    highscore, highscores.get_highscores(settings=highscore)
                )
            except Exception:
                logger.exception("Error getting highscores")
            else:
                if new_best:
                    self.open_highscores_window(highscore, new_best)
                    # try:
                    #     save_highscore_file(
                    #         highscore, self._mf_widget.get_mouse_events(),
                    #     )
                    # except IOError:
                    #     logger.exception("Error saving highscore to file")

    def _open_save_board_modal(self) -> None:
        if not (
            self._state.game_status.finished() or self._state.ui_mode is UIMode.CREATE
        ):
            # TODO: The menubar option should be disabled.
            _msg_popup(
                self,
                QMessageBox.Warning,
                "Save failed",
                "Only able to save boards for finished games, or those created "
                "in 'create' mode.",
            )
            return
        file, _ = QFileDialog.getSaveFileName(
            self,
            caption="Save board",
            directory=str(paths.BOARDS_DIR),
            filter="Minegauler boards (*.mgb)",
        )
        if not file:
            return
        elif not file.endswith(".mgb"):
            file += ".mgb"
        logger.info("Board requested to be saved at %s", file)
        try:
            self._ctrlr.save_current_minefield(file)
        except Exception:
            logger.exception("Error occurred trying to save minefield to file")
            # TODO: Pop up an error message.

    def _open_load_board_modal(self) -> None:
        file, _ = QFileDialog.getOpenFileName(
            self,
            caption="Load board",
            directory=str(paths.BOARDS_DIR),
            filter="Minegauler boards (*.mgb)",
        )
        if not file:
            return
        logger.info("Board requested to be loaded from %s", file)
        try:
            self._ctrlr.load_minefield(file)
        except Exception:
            # TODO: Pop up error message.
            logger.exception("Error occurred trying to load minefield from file")

    def _open_current_info_modal(self) -> None:
        """Open a popup giving the current game info."""
        info = self._ctrlr.get_game_info()
        _CurrentInfoModal(self, info, self._state).show()

    def _open_custom_board_modal(self) -> None:
        """Open the modal popup to select the custom difficulty."""
        if self._state.game_mode is GameMode.REGULAR:
            _CustomBoardModal(
                self,
                self._state.x_size,
                self._state.y_size,
                self._state.mines,
                self._ctrlr.resize_board,
            ).show()
        elif self._state.game_mode is GameMode.SPLIT_CELL:
            _CustomBoardModal(
                self,
                self._state.x_size // 2,
                self._state.y_size // 2,
                self._state.mines,
                lambda x, y, m: self._ctrlr.resize_board(x * 2, y * 2, m),
            ).show()

    def _open_zoom_modal(self) -> None:
        """Open the popup to set the button size."""
        _ButtonSizeModal(self, self._state.btn_size, self.set_button_size).show()

    def _open_text_popup(self, title: str, file: PathLike):
        """Open a text popup window."""
        if title in self._open_subwindows:
            try:
                self._open_subwindows[title].close()
                # self._open_subwindows[title].setFocus()  # TODO: doesn't work!
                # return
            except Exception:
                self._open_subwindows.pop(title)
        win = _TextPopup(self, title, file)
        win.show()
        self._open_subwindows[title] = win

    def _open_retrieve_highscores_modal(self):
        """Open a window to select a highscores file to read in."""
        accepted = False

        def accept_cb():
            nonlocal accepted
            accepted = True

        logger.debug("Opening window to retrieve highscores")
        dialog = QFileDialog(
            parent=self,
            caption="Retrieve highscores",
            directory=str(pathlib.Path.home()),
            filter="highscores.db",
        )
        dialog.setFileMode(QFileDialog.Directory)
        dialog.accepted.connect(accept_cb)
        dialog.exec_()

        if not accepted:
            return  # cancelled

        path = dialog.selectedFiles()[0]
        logger.debug("Selected directory: %s", path)

        path = pathlib.Path(path)
        if not path.exists():
            _msg_popup(
                self,
                QMessageBox.Critical,
                "Not found",
                "Unable to access selected folder.",
            )
            return

        tail = pathlib.Path()
        for part in reversed(["minegauler", "data", "highscores.db"]):
            tail = part / tail
            if (path / tail).is_file():
                file = path / tail
                break
        else:
            _msg_popup(
                self,
                QMessageBox.Warning,
                "Not found",
                "No highscores database found when searching along the path "
                "'.../minegauler/data/highscores.db'. Contact "
                "minegauler@gmail.com if this error is unexpected.",
            )
            return

        try:
            logger.info("Fetching highscores from %s", file)
            try:
                added = highscores.retrieve_highscores(file)
            except highscores.HighscoreReadError as e:
                _msg_popup(
                    self,
                    QMessageBox.Warning,
                    "Highscores retrieve failed",
                    f"Unable to retrieve highscores, error: {e}\n",
                )
            else:
                _msg_popup(
                    self,
                    QMessageBox.Information,
                    "Highscores retrieved",
                    f"Number of highscores added: {added}",
                )
        except Exception as e:
            _msg_popup(
                self,
                QMessageBox.Critical,
                "Error",
                str(e) + " Contact minegauler@gmail.com if this error is unexpected.",
            )

    def _open_play_highscore_modal(self):
        """Open a modal window to select a highscore file."""
        logger.debug("Opening window to select a highscore to play back")
        hs_file, _ = QFileDialog.getOpenFileName(
            parent=self,
            caption="Play highscore",
            directory=str(paths.DATA_DIR),
            filter="Minegauler highscores (*.mgh)",
        )
        logger.debug("Selected file: %s", hs_file)
        if not hs_file:
            return  # cancelled

        hs_file = pathlib.Path(hs_file)
        try:
            hs, cell_updates = utils.read_highscore_file(hs_file)
            x_size, y_size, _ = hs.difficulty.get_board_values()
            win = simulate.SimulationMinefieldWidget(self, x_size, y_size, cell_updates)
        except Exception as e:
            logger.exception("Error reading highscore file")
            _msg_popup(
                self, QMessageBox.Warning, "Error loading highscore file", str(e)
            )
        else:
            win.show()

    def _open_advanced_opts_modal(self):
        _AdvancedOptionsModal(self, self).show()

    def get_gui_opts(self) -> GUIOptsStruct:
        return GUIOptsStruct.from_structs(self._state, self._state.pending_game_state)

    def open_highscores_window(
        self,
        settings: Optional[HighscoreSettingsStruct] = None,
        sort_by: Optional[str] = None,
    ) -> None:
        """
        Open the highscores window.

        :param settings:
            Optionally specify the highscore settings to display highscores
            for. By default the current frontend state is used.
        :param sort_by:
            Optionally specify the key to sort the highscores by. Defaults to
            the previous sort key.
        """
        if self._state.difficulty is Difficulty.CUSTOM:
            _msg_popup(
                self,
                QMessageBox.Warning,
                "Highscores unavailable",
                "No highscores available for the active settings.\n"
                "Highscores are only stored for the standard board difficulties, "
                "and currently only for the 'regular' game mode.",
            )
            return
        if self._open_subwindows.get("highscores"):
            self._open_subwindows.get("highscores").close()
        if not settings:
            settings = HighscoreSettingsStruct(
                game_mode=self._state.game_mode,
                difficulty=self._state.difficulty,
                per_cell=self._state.per_cell,
                drag_select=self._state.drag_select,
            )
        if sort_by:
            self._state.highscores_state.sort_by = sort_by
        self._state.highscores_state.name_hint = self._state.name
        win = highscores_ui.HighscoresWindow(
            self, settings, self._state.highscores_state
        )
        win.show()
        self._open_subwindows["highscores"] = win

    def set_button_size(self, size: int) -> None:
        """Set the cell size in pixels."""
        self._state.btn_size = size
        self._mf_widget.update_btn_size(size)

    def factory_reset(self) -> None:
        """
        Reset to original state.
        """
        self._ctrlr.reset_settings()
        self._state.reset()
        for img_group in CellImageType:
            if img_group is not CellImageType.ALL:
                self._change_styles({img_group: self._state.styles[img_group]})
        self._name_entry_widget.setText("")


class _CurrentInfoModal(QDialog):
    """A popup window with information about the current game."""

    def __init__(self, parent: QWidget, info: api.GameInfo, state_: state.State):
        super().__init__(parent)
        self._info = info
        self._state = state_
        self.setWindowTitle("Game info")
        self.setModal(True)
        self._setup_ui()

    def _setup_ui(self) -> None:
        """Set up the UI."""
        base_layout = QVBoxLayout(self)
        base_layout.addWidget(QLabel("Information about the current game:", self))
        info = self._info
        info_text = textwrap.dedent(
            f"""\
            {info.x_size} x {info.y_size} grid
            {info.mines} mines
            Max mines per cell: {info.per_cell}
            Drag-select: {self._state.drag_select}
            """
        )
        if info.game_state.finished():
            assert info.started_info is not None
            fin_info = info.started_info
            start_time = format_timestamp(fin_info.start_time)
            state = info.game_state.name.capitalize()
            info_text += textwrap.dedent(
                f"""\

                Game was started at {start_time}
                {state} after {fin_info.elapsed + 0.005:.2f} seconds
                The board was {fin_info.prop_flagging * 100:.1f}% flagged
                The 3bv was: {fin_info.bbbv}
                The 3bv/s rate was: {fin_info.bbbvps + 0.005:.2f}
                """
            )
            if info.game_state is GameState.LOST:
                info_text += textwrap.dedent(
                    f"""\

                    Remaining 3bv: {fin_info.rem_bbbv}
                    Game was {fin_info.prop_complete * 100:.1f}% complete
                    """
                )
                if fin_info.prop_complete > 0:
                    info_text += "Predicted completion time: {:.2f} seconds".format(
                        fin_info.elapsed / fin_info.prop_complete
                    )
        base_layout.addWidget(QLabel(info_text, self))
        ok_btn = QPushButton("Ok", self)
        ok_btn.pressed.connect(self.close)
        base_layout.addWidget(ok_btn)


class _CustomBoardModal(QDialog):
    """A popup window to select custom board dimensions."""

    def __init__(
        self,
        parent: QWidget,
        cols: int,
        rows: int,
        mines: int,
        callback: Callable[[int, int, int], None],
    ):
        super().__init__(parent)
        self._cols: int = cols
        self._rows: int = rows
        self._mines: int = mines
        self._callback: Callable[[int, int, int], None] = callback
        self.setWindowTitle("Custom")
        self.setModal(True)
        self._setup_ui()

    def _setup_ui(self) -> None:
        """Set up the UI."""
        base_layout = QVBoxLayout(self)
        base_layout.addWidget(
            QLabel("Select the number of rows, columns and mines.", self)
        )

        # Set up the input sliders and spin boxes.
        grid_layout = QGridLayout()
        base_layout.addLayout(grid_layout)

        def make_row(
            row: int, label: str, current: int, minval: int, maxval: int
        ) -> _SliderSpinner:
            grid_layout.addWidget(QLabel(label, self), row, 0)
            slider = _SliderSpinner(self)
            grid_layout.addWidget(slider, row, 1)
            slider.setValue(current)
            slider.setMinimum(minval)
            slider.setMaximum(maxval)
            return slider

        y_slider = make_row(1, "Rows", self._rows, 2, 50)
        x_slider = make_row(2, "Columns", self._cols, 2, 50)
        m_slider = make_row(3, "Mines", self._mines, 1, self._rows * self._cols - 1)

        def set_mines_max():
            m_slider.setMaximum(x_slider.value() * y_slider.value() - 1)

        x_slider.valueChanged.connect(set_mines_max)
        y_slider.valueChanged.connect(set_mines_max)

        # Ok/Cancel buttons.
        def ok_pressed():
            self._callback(x_slider.value(), y_slider.value(), m_slider.value())
            self.close()

        ok_btn = QPushButton("Ok", self)
        ok_btn.pressed.connect(ok_pressed)
        cancel_btn = QPushButton("Cancel", self)
        cancel_btn.pressed.connect(self.close)
        btns_layout = QHBoxLayout()
        base_layout.addLayout(btns_layout)
        btns_layout.addWidget(ok_btn)
        btns_layout.addWidget(cancel_btn)


class _ButtonSizeModal(QDialog):
    """A popup window to select the button size."""

    def __init__(self, parent: QWidget, btn_size: int, callback: Callable):
        super().__init__(parent)
        self.setWindowTitle("Button size")
        self._btn_size: int = btn_size
        self._callback: Callable = callback
        self.setModal(True)
        self._setup_ui()

    def _setup_ui(self) -> None:
        """Set up the UI."""
        base_layout = QVBoxLayout(self)
        base_layout.addWidget(QLabel("Select the button size in pixels.", self))

        # Set up the input sliders and spin boxes.
        slider = _SliderSpinner(self)
        slider.setValue(self._btn_size)
        slider.setMinimum(16)
        slider.setMaximum(40)
        base_layout.addWidget(slider)

        # Ok/Cancel buttons.
        def ok_pressed():
            self._callback(slider.value())
            self.close()

        ok_btn = QPushButton("Ok", self)
        ok_btn.pressed.connect(ok_pressed)
        cancel_btn = QPushButton("Cancel", self)
        cancel_btn.pressed.connect(self.close)
        btns_layout = QHBoxLayout()
        base_layout.addLayout(btns_layout)
        btns_layout.addWidget(ok_btn)
        btns_layout.addWidget(cancel_btn)


class _AdvancedOptionsModal(QDialog):
    """A popup window to select advanced options."""

    def __init__(self, parent: QWidget, main_window: MinegaulerGUI):
        super().__init__(parent)
        self.setWindowTitle("Advanced options")
        self.setModal(True)
        self._main_window = main_window
        self._setup_ui()

    def _setup_ui(self):
        """Set up the window layout."""
        vlayout = QVBoxLayout(self)
        vlayout.addWidget(
            QLabel("These options are in beta - stability not guaranteed", self)
        )

        def add_opt(label: str, func: Callable):
            hlayout = QHBoxLayout(self)
            vlayout.addLayout(hlayout)
            hlayout.addWidget(QLabel(label, self))
            button = QPushButton("Enable", self)
            button.clicked.connect(func)
            button.clicked.connect(lambda: button.setDisabled(True))
            hlayout.addWidget(button)

        add_opt("Enable main window maximising", self._enable_maximise)

    def _enable_maximise(self):
        """Enable maximising the main window."""
        win = self._main_window

        def update_size():
            win.setMinimumSize(win.minimumSizeHint())
            if not win.isMaximized():
                win.resize(win.sizeHint())
                win.adjustSize()

        win.hide()
        win.setWindowFlag(Qt.WindowMaximizeButtonHint)
        win.setMaximumSize(QWIDGETSIZE_MAX, QWIDGETSIZE_MAX)
        win._mf_widget.size_changed.disconnect(win._update_size)
        win._update_size = update_size
        win._mf_widget.size_changed.connect(win._update_size)
        win.show()


class _SliderSpinner(QWidget):
    """A combination of a slider and a spinbox."""

    valueChanged = pyqtSignal(int)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.slider = QSlider(Qt.Horizontal)
        self.numbox = QSpinBox()
        self.numbox.setRange(self.slider.minimum(), self.slider.maximum())
        self.numbox.setFixedWidth(50)
        self.slider.valueChanged.connect(self.numbox.setValue)
        self.slider.rangeChanged.connect(self.numbox.setRange)
        self.numbox.valueChanged.connect(self.slider.setValue)
        self.slider.valueChanged.connect(self.valueChanged.emit)
        self.valueChanged.connect(self.setValue)
        layout = QHBoxLayout(self)
        layout.addWidget(self.slider)
        layout.addWidget(self.numbox)

    def setMinimum(self, minval: int):
        self.slider.setMinimum(minval)

    def setMaximum(self, maxval: int):
        self.slider.setMaximum(maxval)

    def setValue(self, value: int):
        self.slider.setValue(value)

    def value(self) -> int:
        return self.slider.value()


class _NameEntryBar(QLineEdit):
    """Entry bar for entering a name."""

    name_updated_signal = pyqtSignal(str)

    def __init__(self, parent, text: str = ""):
        super().__init__(parent)
        self.setText(text)
        self.setPlaceholderText("Enter name here")
        self.setAlignment(Qt.AlignCenter)
        font = QFont("Helvetica")
        font.setBold(True)
        self.setFont(font)

    def keyPressEvent(self, event: QKeyEvent):
        super().keyPressEvent(event)
        if event.key() in [Qt.Key_Return, Qt.Key_Enter]:
            self.clearFocus()

    def focusOutEvent(self, event: QFocusEvent):
        super().focusOutEvent(event)
        self.setText(self.text().strip())
        logger.info("Setting name to %r", self.text())
        self.name_updated_signal.emit(self.text())


class _TextPopup(QWidget):
    """A popup window containing a block of text."""

    def __init__(self, parent: Optional[QWidget], title: str, file: PathLike):
        super().__init__(parent)
        self.setWindowFlag(Qt.Window)
        self.setWindowTitle(title)
        self.setMinimumSize(200, 100)

        self._text_widget = QLabel(self)
        self._ok_button = QPushButton(self)
        self._setup_ui()

        with open(file, mode="r", encoding="utf-8") as f:
            self._text_widget.setText(f.read())

    def _setup_ui(self) -> None:
        """Set up the UI layout."""
        lyt = QVBoxLayout(self)
        # lyt.setAlignment(Qt.AlignCenter)

        self._text_widget.setLineWidth(1)
        self._text_widget.setFrameShape(QFrame.Panel)
        self._text_widget.setMargin(10)
        self._text_widget.setStyleSheet(
            """
            background-color: white;
            """
        )
        self._text_widget.setWordWrap(True)
        lyt.addWidget(self._text_widget)

        self._ok_button.setText("Ok")
        self._ok_button.setMaximumWidth(50)
        self._ok_button.clicked.connect(self.close)
        lyt.addWidget(self._ok_button)

    def sizeHint(self) -> QSize:
        size = super().sizeHint()
        area = size.width() * size.height()
        return QSize(int(area**0.5), int(area**0.4))



# February 2022, Lewis Gaul

"""
Compatibility with old highscore formats.

The submodule implementations are expected to never change, to maintain
compatibility with old versions of the app. For this reason there should also
be no dependency on the rest of the highscores package, which is free to evolve
alongside the app.

"""

__all__ = ("ConversionFunc", "HighscoreReadError", "read_highscores")

import pathlib
import sqlite3
from typing import Callable, Iterable

from ...shared.types import PathLike
from ..base import HighscoreStruct
from . import sqlite_v0, sqlite_v1


ConversionFunc = Callable[[PathLike], Iterable[HighscoreStruct]]


class HighscoreReadError(Exception):
    pass


def read_highscores(path: PathLike) -> Iterable[HighscoreStruct]:
    """
    Read highscores from the given path.

    Delegates to the appropriate submodule's implementation for older versions.

    :param path:
        Path to highscores, may be file or directory as required.
    :return:
        Iterable of retrieved highscores.
    :raise HighscoreReadError:
        If unable to read highscores from the given path.
    """
    path = pathlib.Path(path)
    if (path / "data" / "highscores.db").is_file():
        path = path / "data" / "highscores.db"
    elif (path / "highscores.db").is_file():
        path = path / "highscores.db"

    if path.suffix == ".db":
        with sqlite3.connect(str(path)) as conn:
            sqlite_db_version = next(conn.execute("PRAGMA user_version"))[0]
        func: ConversionFunc
        if sqlite_db_version == 0:
            func = sqlite_v0.read_highscores
        elif sqlite_db_version == 1:
            func = sqlite_v1.read_highscores
        else:
            raise HighscoreReadError(
                f"Unrecognised SQLite DB version '{sqlite_db_version}'"
            )
        try:
            return func(path)
        except Exception as e:
            raise HighscoreReadError(
                f"Failed to read v{sqlite_db_version} SQLite DB"
            ) from e
    else:
        raise HighscoreReadError(f"Unrecognised DB extension '{path.suffix}'")



# December 2019, Lewis Gaul

"""
Code not belonging to another specific sub-package.

"""

__all__ = (
    "AllOptsStruct",
    "GameOptsStruct",
    "GUIOptsStruct",
    "read_settings_from_file",
    "write_settings_to_file",
)

from .utils import (
    AllOptsStruct,
    GameOptsStruct,
    GUIOptsStruct,
    read_settings_from_file,
    write_settings_to_file,
)



# February 2020, Lewis Gaul

"""
Bot utilities.

"""

__all__ = (
    "BOT_NAME",
    "NO_TAG_USERS",
    "USER_NAMES",
    "Matchup",
    "PlayerInfo",
    "get_highscores",
    "get_matchups",
    "get_message",
    "get_highscore_times",
    "get_player_info",
    "is_highscore_new_best",
    "send_group_message",
    "send_message",
    "send_new_best_message",
    "send_myself_message",
    "set_bot_access_token",
    "set_user_nickname",
    "user_from_email",
)

import collections
import json
import logging
import os
import pathlib
import sys
from typing import Iterable, List, Optional, Tuple

import requests
from requests_toolbelt import MultipartEncoder

from minegauler.app import highscores as hs
from minegauler.app.shared.types import Difficulty, GameMode


logger = logging.getLogger(__name__)


USER_NAMES = dict()
if hasattr(sys, "frozen") and hasattr(sys, "_MEIPASS"):  # in pyinstaller EXE
    _USER_NAMES_FILE = pathlib.Path(__file__).parents[2] / "bot/users.json"
    os.makedirs(_USER_NAMES_FILE.parent, exist_ok=True)
else:
    _USER_NAMES_FILE = pathlib.Path(__file__).parent / "users.json"
NO_TAG_USERS = {"_felix", "_kunz", "esinghal", "richcoop"}

BOT_NAME = "minegaulerbot"
_BOT_ACCESS_TOKEN = ""
_BOT_EMAIL = f"{BOT_NAME}@webex.bot"
_WEBEX_GROUP_ROOM_ID = (
    "Y2lzY29zcGFyazovL3VzL1JPT00vNzYyNjI4NTAtMzg3Ni0xMWVhLTlhM2ItODMyNzMyZDlkZTg3"
)

_API_BASEURL = "http://minegauler.lewisgaul.co.uk/api/v1/highscores"


# ------------------------------------------------------------------------------
# External
# ------------------------------------------------------------------------------


def set_bot_access_token(token: str) -> None:
    global _BOT_ACCESS_TOKEN
    _BOT_ACCESS_TOKEN = token


def get_message(msg_id: str) -> str:
    response = requests.get(
        f"https://api.ciscospark.com/v1/messages/{msg_id}",
        headers={"Authorization": f"Bearer {_BOT_ACCESS_TOKEN}"},
    )
    response.raise_for_status()
    return response.json()["text"]


def send_message(
    room_id: str, text: str, *, is_person_id=False, markdown=False
) -> requests.Response:
    logger.debug(
        "Sending message to %s:\n%s", "person" if is_person_id else "room", text
    )
    id_field = "toPersonId" if is_person_id else "roomId"
    text_field = "markdown" if markdown else "text"
    multipart = MultipartEncoder({text_field: text, id_field: room_id})
    response = requests.post(
        "https://api.ciscospark.com/v1/messages",
        data=multipart,
        headers={
            "Authorization": f"Bearer {_BOT_ACCESS_TOKEN}",
            "Content-Type": multipart.content_type,
        },
    )
    response.raise_for_status()
    return response


def send_myself_message(text: str) -> requests.Response:
    return send_message(_get_my_id(), text, is_person_id=True)


def send_group_message(text: str) -> requests.Response:
    return send_message(_WEBEX_GROUP_ROOM_ID, text)


def send_new_best_message(h: hs.HighscoreStruct) -> None:
    """Send a group message when a new personal time record is set."""
    diff = h.difficulty.name.lower()
    drag_select = "on" if h.drag_select else "off"
    if h.game_mode is GameMode.REGULAR:
        send_group_message(
            f"New personal record of {h.elapsed:.2f} set by {h.name} on {diff}!\n"
            f"Settings: drag-select={drag_select}, per-cell={h.per_cell}"
        )
    else:
        assert h.game_mode is GameMode.SPLIT_CELL
        send_group_message(
            f"New personal split-cell record of {h.elapsed:.2f} set by {h.name} on {diff}!\n"
            f"Settings: drag-select={drag_select}, per-cell={h.per_cell}"
        )


def read_users_file():
    global USER_NAMES
    try:
        with open(_USER_NAMES_FILE) as f:
            USER_NAMES = json.load(f)
    except FileNotFoundError:
        logger.info("%s file not found", _USER_NAMES_FILE)


def save_users_file():
    with open(_USER_NAMES_FILE, "w") as f:
        json.dump(USER_NAMES, f)


def user_from_email(email: str) -> str:
    return email.split("@", maxsplit=1)[0]


def user_to_email(user: str) -> str:
    return f"{user}@cisco.com"


def tag_user(user: str) -> str:
    if user.strip() and user not in NO_TAG_USERS:
        return f"<@personEmail:{user_to_email(user)}|{user}>"
    else:
        return user


def set_user_nickname(user: str, nickname: str) -> None:
    USER_NAMES[user] = nickname
    save_users_file()


def get_highscore_times(
    difficulty: Optional[Difficulty],
    *,
    game_mode: GameMode = GameMode.REGULAR,
    drag_select: Optional[bool] = None,
    per_cell: Optional[int] = None,
    users: Optional[Iterable[str]] = None,
) -> List[Tuple[str, float]]:
    if difficulty is Difficulty.CUSTOM:
        raise ValueError("No highscores for custom difficulty")
    if users is None:
        users = USER_NAMES.values()
    lower_users = {u.lower(): u for u in users}

    if difficulty:
        highscores = hs.filter_and_sort(
            get_highscores(
                game_mode=game_mode,
                difficulty=difficulty,
                drag_select=drag_select,
                per_cell=per_cell,
            )
        )
        times = {
            lower_users[h.name.lower()]: h.elapsed
            for h in highscores
            if h.name.lower() in lower_users
        }
    else:
        times = {
            u: _get_combined_highscore(
                u, game_mode=game_mode, drag_select=drag_select, per_cell=per_cell
            )
            for u in users
        }

    return sorted(times.items(), key=lambda x: x[1])


Matchup = collections.namedtuple("Matchup", "user1, time1, user2, time2, percent")


def get_matchups(
    times: Iterable[Tuple[str, float]], include_users: Optional[Iterable[str]] = None
) -> List[Matchup]:
    times = sorted(times, key=lambda x: x[1], reverse=True)
    matchups = set()
    while times:
        # Avoid repeating matchups or comparing users against themselves.
        user1, time1 = times.pop()
        for user2, time2 in times:
            if (
                include_users
                and user1 not in include_users
                and user2 not in include_users
            ):
                continue
            assert time2 >= time1
            percent = 100 * (time2 - time1) / time1
            matchups.add(Matchup(user1, time1, user2, time2, percent))

    return sorted(matchups, key=lambda x: x.percent)


PlayerInfo = collections.namedtuple(
    "PlayerInfo", "username, nickname, combined_time, types_beaten, last_highscore"
)


def get_player_info(username: str) -> PlayerInfo:
    name = USER_NAMES[username]
    highscores = [h for m in GameMode for h in get_highscores(game_mode=m, name=name)]
    combined_time = _get_combined_highscore(name)
    last_highscore = max(h.timestamp for h in highscores) if highscores else None
    hs_types = len(
        {
            (h.game_mode, h.difficulty.lower(), h.drag_select, h.per_cell)
            for h in highscores
        }
    )
    return PlayerInfo(username, name, combined_time, hs_types, last_highscore)


def is_highscore_new_best(h: hs.HighscoreStruct) -> Optional[str]:
    all_highscores = get_highscores(settings=h, name=h.name)
    return hs.is_highscore_new_best(h, all_highscores)


def get_highscores(
    *,
    settings: Optional[hs.HighscoreSettingsStruct] = None,
    game_mode: GameMode = GameMode.REGULAR,
    difficulty: Optional[Difficulty] = None,
    per_cell: Optional[int] = None,
    drag_select: Optional[bool] = None,
    name: Optional[str] = None,
) -> Iterable[hs.HighscoreStruct]:
    """
    Get highscores using the REST API.

    :param settings:
        Highscore filter to apply.
    :param game_mode:
        The game mode to get highscores for. Ignored if settings given.
    :param difficulty:
        Optionally specify difficulty to filter by. Ignored if settings given.
    :param per_cell:
        Optionally specify per-cell to filter by. Ignored if settings given.
    :param drag_select:
        Optionally specify drag-select to filter by. Ignored if settings given.
    :param name:
        Optionally specify a name to filter by.
    :raises Exception:
        If the HTTP request fails or returns bad data.
    :return:
        Matching highscores.
    """
    if settings is not None:
        game_mode = settings.game_mode
        difficulty = settings.difficulty
        per_cell = settings.per_cell
        drag_select = settings.drag_select
    args = [f"game_mode={game_mode.name.lower()}"]
    if difficulty is not None:
        args.append(f"difficulty={difficulty.name[0]}")
    if per_cell is not None:
        args.append(f"per_cell={per_cell}")
    if drag_select is not None:
        args.append(f"drag_select={int(drag_select)}")
    if name is not None:
        args.append(f"name={name}")
    url = _API_BASEURL + "?" + "&".join(args)
    response = requests.get(url)
    return [hs.HighscoreStruct.from_dict(h) for h in response.json()]


# ------------------------------------------------------------------------------
# Internal
# ------------------------------------------------------------------------------


def _strbool(b: bool) -> str:
    return "True" if b else "False"


def _get_combined_highscore(
    name: str,
    *,
    game_mode: GameMode = GameMode.REGULAR,
    per_cell: Optional[int] = None,
    drag_select: Optional[bool] = None,
) -> float:
    total = 0
    for diff in ["b", "i", "e"]:
        all_highscores = get_highscores(
            game_mode=game_mode,
            name=name,
            drag_select=drag_select,
            per_cell=per_cell,
        )
        highscores = [h.elapsed for h in all_highscores if h.difficulty.lower() == diff]
        total += min(highscores) if highscores else 1000
    return total


def _get_person_id(name_or_email: str) -> str:
    if "@" in name_or_email:
        params = {"email": name_or_email}
    else:
        params = {"displayName": name_or_email}
    response = requests.get(
        f"https://api.ciscospark.com/v1/people",
        params=params,
        headers={"Authorization": f"Bearer {_BOT_ACCESS_TOKEN}"},
    )
    response.raise_for_status()
    return response.json()["items"][0]["id"]


def _get_my_id() -> str:
    return _get_person_id("Lewis Gaul")


def _get_bot_id() -> str:
    return _get_person_id("Minegauler Bot")



# February 2020, Lewis Gaul

"""
Definition of bot HTTP routes.

"""

__all__ = ("activate_bot_msg_handling", "init_route_handling", "new_highscore_hook")

import logging
import os
import re
import sys

import flask
import requests

from minegauler import server
from minegauler.app import highscores as hs

from . import msgparse, utils


logger = logging.getLogger(__name__)


# ------------------------------------------------------------------------------
# REST API
# ------------------------------------------------------------------------------


def bot_message():
    """Receive a notification of a bot message."""
    data = flask.request.get_json()["data"]
    logger.debug("POST bot message: %s", data)
    user = utils.user_from_email(data["personEmail"])
    send_welcome = False
    if user == utils.BOT_NAME:
        # Ignore messages sent by the bot.
        return "", 200
    elif user not in utils.USER_NAMES:
        # If the user is not yet tracked, add the username as the default name.
        logger.debug("Adding new user %r", user)
        utils.set_user_nickname(user, user)
        send_welcome = True

    msg_id = data["id"]
    try:
        msg_text = utils.get_message(msg_id)
    except requests.HTTPError:
        logger.exception(f"Error getting message from {user}")
        _send_myself_error_msg("getting message")
        raise

    logger.debug("Fetched message content: %r", msg_text)
    msg = re.sub(r"@?Minegauler(?: Bot)?", "", msg_text, 1).strip()
    logger.debug("Handling message: %r", msg)

    if not msg:
        return "", 200

    room_id = data["roomId"]
    person_id = data["personId"]
    room_type = msgparse.RoomType(data["roomType"])

    if send_welcome:
        try:
            utils.send_message(
                person_id, msgparse.GENERAL_INFO, is_person_id=True, markdown=True
            )
        except requests.HTTPError:
            logger.exception("Error sending bot welcome message")
            _send_myself_error_msg("sending bot welcome message")

    send_to_person_id = False
    send_to_id = room_id
    try:
        resp_msg = msgparse.parse_msg(
            msg, room_type, allow_markdown=True, username=user
        )
    except msgparse.InvalidArgsError as e:
        resp_msg = str(e)
        if room_type is msgparse.RoomType.GROUP:
            # Send error message to direct chat.
            send_to_person_id = True
            send_to_id = person_id

    try:
        utils.send_message(
            send_to_id, resp_msg, is_person_id=send_to_person_id, markdown=True
        )
    except requests.HTTPError:
        logger.exception("Error sending bot response message")
        _send_myself_error_msg("sending bot response message")

    return "", 200


def new_highscore_hook(highscore: hs.HighscoreStruct) -> None:
    if highscore.name != "Siwel G":
        try:
            utils.send_myself_message(f"New highscore added:\n{highscore}")
        except Exception:
            logger.exception("Error sending webex message")

    if (
        highscore.name in utils.USER_NAMES.values()
        and utils.is_highscore_new_best(highscore) == "time"
    ):
        try:
            utils.send_new_best_message(highscore)
        except Exception:
            logger.exception("Error sending webex message for new best")
            _send_myself_error_msg("sending webex message for new best")


def activate_bot_msg_handling(app: flask.app.Flask) -> None:
    """Register a route to handle bot messages."""

    def bot_message_with_error_catching():
        try:
            return bot_message()
        except Exception:
            logger.exception("Unexpected error occurred handling a bot message")
            _send_myself_error_msg("<uncaught> occurred")
            raise

    app.add_url_rule(
        "/bot/message", "bot_message", bot_message_with_error_catching, methods=["POST"]
    )


def _send_myself_error_msg(error: str) -> None:
    try:
        utils.send_myself_message(f"Error {error}, see server logs")
    except requests.HTTPError:
        logger.exception("Error sending myself bot message when handling error")


def init_route_handling(app: flask.app.Flask):
    if "BOT_ACCESS_TOKEN" not in os.environ:
        logger.error("No 'BOT_ACCESS_TOKEN' env var set")
        sys.exit(1)
    utils.set_bot_access_token(os.environ["BOT_ACCESS_TOKEN"])

    utils.read_users_file()

    activate_bot_msg_handling(app)

    server.add_new_highscore_hook(new_highscore_hook)



# December 2018, Lewis Gaul

"""
Utilities for the frontend.

Exports
-------
.. class:: CellUpdate_T
    A cell update type alias.

.. class:: MouseMove
    A mouse move tuple.

.. function:: save_highscore_file
    Save a highscore to file.

.. function:: read_highscore_file
    Read data from a highscore file.

"""

__all__ = (
    "CellUpdate_T",
    "MouseMove",
    "read_highscore_file",
    "save_highscore_file",
)

import gzip
import json
import logging
import pathlib
from collections import namedtuple
from typing import Iterable, List, Mapping, Tuple

import attr

from .. import paths
from ..highscores import HighscoreStruct
from ..shared.types import CellContents, Coord, PathLike
from ..shared.utils import format_timestamp


logger = logging.getLogger(__name__)

CellUpdate_T = Tuple[float, Mapping[Coord, CellContents]]


# TODO
class MouseMove(namedtuple("_MouseMove", ["elapsed", "position"])):
    """A mouse move tuple."""

    def __new__(cls, elapsed: float, position: Coord):
        return super().__new__(cls, elapsed, position)


def save_highscore_file(
    highscore: HighscoreStruct, cell_updates: Iterable[CellUpdate_T]
) -> pathlib.Path:
    """
    Save a highscore to file.

    :param highscore:
        The highscore to save.
    :param cell_updates:
        The cell updates that were made during the game.
    :return:
        The path the file is saved at.
    """
    fname = (
        "{0.name}_{0.difficulty.value}_{0.elapsed:.2f}_{0.bbbv}_"
        "max={0.per_cell}_drag={0.drag_select}_{1}.mgh"
    ).format(
        highscore,
        format_timestamp(highscore.timestamp).replace(" ", "_").replace(":", "-"),
    )
    data = {
        "highscore": attr.asdict(highscore),
        "cell_updates": [
            (t, [(c, str(x)) for c, x in updates.items()])
            for t, updates in cell_updates
        ],
    }
    paths.DATA_DIR.mkdir(exist_ok=True)
    with gzip.open(paths.DATA_DIR / fname, "wt") as f:
        json.dump(data, f)
    return paths.DATA_DIR / fname


def read_highscore_file(
    path: PathLike,
) -> Tuple[HighscoreStruct, List[CellUpdate_T]]:
    """
    Read data from a highscore file.

    :param path:
        Path to the file.
    :return:
        The highscore struct and a list of cell updates.
    """
    with gzip.open(path, "rt") as f:
        data = json.load(f)
    return HighscoreStruct(**data["highscore"]), data["cell_updates"]



# January 2020, Lewis Gaul

import runpy
import sys

import setuptools


with open("README.pypi.md", "r", encoding="utf-8") as f:
    long_description = f.read()

version = runpy.run_path("minegauler/app/_version.py")["__version__"]
# Should be a release version, e.g. "4.0.1"
if "-" in version:
    print("WARNING: Version is not a release version", file=sys.stderr)


setuptools.setup(
    name="minegauler",
    version=version,
    author="Lewis Gaul",
    author_email="minegauler@gmail.com",
    description="A clone of the original minesweeper game with many added features",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/LewisGaul/minegauler",
    project_urls={
        "Bug Reports": "https://github.com/LewisGaul/minegauler/issues",
        "Source": "https://github.com/LewisGaul/minegauler/",
        "Background": "https://www.lewisgaul.co.uk/blog/coding/2020/02/12/minegauler/",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta",
        "Intended Audience :: End Users/Desktop",
        "Natural Language :: English",
        "Topic :: Games/Entertainment :: Puzzle Games",
    ],
    python_requires=">=3.7",
    packages=[
        "minegauler",
        *[
            "minegauler." + subpkg
            for subpkg in (
                *setuptools.find_packages(
                    where="minegauler/", include=("app", "app.*", "bot")
                ),
            )
        ],
    ],
    package_data={
        "minegauler.app": [
            "boards/sample.mgb",
            "files/*.txt",
            "images/icon.ico",
            "images/faces/*",
            "images/buttons/*/*",
            "images/markers/*/*",
            "images/numbers/*/*",
        ],
    },
    install_requires=[
        "attrs",
        "PyQt5",
        "requests",
        "typing-extensions",
    ],
    extras_require={
        "bot": [
            "pytz",
            "requests-toolbelt",
            "tabulate",
        ],
    },
    entry_points={
        "gui_scripts": [
            "minegauler = minegauler.app.__main__:main",
        ],
        "console_scripts": [
            "minegauler-bot = minegauler.bot.msgparse:main [bot]",
        ],
    },
    zip_safe=False,
)



# February 2022, Lewis Gaul

"""
Compatibility with v0 SQLite highscore format.

This is the initial SQLite format that comes with v4.0 minegauler.

"""

__all__ = ("read_highscores",)

import sqlite3
from typing import Iterable

from ...shared.types import PathLike
from ..base import HighscoreStruct


_TABLE_NAME = "highscores"


def read_highscores(path: PathLike) -> Iterable[HighscoreStruct]:
    ret = set()
    with sqlite3.connect(str(path)) as conn:
        cursor = conn.execute(f"SELECT * FROM {_TABLE_NAME}")
        for row in cursor:
            # First row entry is 'index' (ignore).
            ret.add(
                HighscoreStruct(
                    game_mode="regular",  # only mode supported
                    difficulty=row[1],
                    per_cell=row[2],
                    drag_select=row[3],
                    name=row[4],
                    timestamp=row[5],
                    elapsed=row[6],
                    bbbv=row[7],
                    bbbvps=row[8],
                    flagging=row[9],
                )
            )
    return ret



# December 2019, Lewis Gaul

"""
State shared by the frontend.

Exports
-------
.. class:: State
    A collection of all frontend state.

.. class:: PerGameState
    State corresponding to a single game.

.. class:: HighscoreWindowState
    State for the highscores window.

"""

__all__ = ("HighscoreWindowState", "PerGameState", "State")

import logging
from typing import Dict, Optional

import attr

from ..highscores import HighscoreStruct
from ..shared.types import CellImageType, Difficulty, GameMode, GameState, UIMode
from ..shared.utils import GameOptsStruct, GUIOptsStruct, StructConstructorMixin


logger = logging.getLogger(__name__)


@attr.attrs(auto_attribs=True)
class PerGameState(StructConstructorMixin):
    """State that applies to an in-progress game."""

    x_size: int = 8
    y_size: int = 8
    mines: int = 10
    difficulty: Difficulty = Difficulty.BEGINNER
    first_success: bool = True
    per_cell: int = 1
    lives: int = 1
    drag_select: bool = False
    mode: GameMode = GameMode.REGULAR


class HighscoreWindowState(StructConstructorMixin):
    """State associated with the highscores window."""

    name_filter: Optional[str] = None
    name_hint: str = ""
    flagging_filter: Optional[str] = None
    sort_by: str = "time"
    current_highscore: Optional[HighscoreStruct] = None


@attr.attrs(auto_attribs=True, kw_only=True)
class State:
    """All state shared between widgets."""

    _current_game_state: PerGameState = PerGameState()
    _pending_game_state: Optional[PerGameState] = None

    btn_size: int = 16
    name: str = ""
    styles: Dict[CellImageType, str] = {
        CellImageType.BUTTONS: "Standard",
        CellImageType.NUMBERS: "Standard",
        CellImageType.MARKERS: "Standard",
    }

    _game_status: GameState = GameState.READY

    ui_mode: UIMode = UIMode.GAME

    highscores_state: HighscoreWindowState = HighscoreWindowState()

    # ---------------------------------
    # Handling for game state fields
    # ---------------------------------
    def _update_game_state(self, field: str, value) -> None:
        if self.game_status is GameState.READY:
            setattr(self._current_game_state, field, value)
        else:
            if not self.has_pending_game_state():
                self._pending_game_state = self._current_game_state.copy()
            setattr(self.pending_game_state, field, value)

    @property
    def x_size(self):
        return self._current_game_state.x_size

    @x_size.setter
    def x_size(self, value):
        self._update_game_state("x_size", value)

    @property
    def pending_x_size(self):
        if self.has_pending_game_state():
            return self.pending_game_state.x_size
        else:
            return self._current_game_state.x_size

    @property
    def y_size(self):
        return self._current_game_state.y_size

    @y_size.setter
    def y_size(self, value):
        self._update_game_state("y_size", value)

    @property
    def pending_y_size(self):
        if self.has_pending_game_state():
            return self.pending_game_state.y_size
        else:
            return self._current_game_state.y_size

    @property
    def mines(self):
        return self._current_game_state.mines

    @mines.setter
    def mines(self, value):
        self._update_game_state("mines", value)

    @property
    def pending_mines(self):
        if self.has_pending_game_state():
            return self.pending_game_state.mines
        else:
            return self._current_game_state.mines

    @property
    def difficulty(self) -> Difficulty:
        return self._current_game_state.difficulty

    @difficulty.setter
    def difficulty(self, value):
        self._update_game_state("difficulty", value)

    @property
    def pending_difficulty(self):
        if self.has_pending_game_state():
            return self.pending_game_state.difficulty
        else:
            return self._current_game_state.difficulty

    @property
    def first_success(self):
        return self._current_game_state.first_success

    @first_success.setter
    def first_success(self, value):
        self._update_game_state("first_success", value)

    @property
    def pending_first_success(self):
        if self.has_pending_game_state():
            return self.pending_game_state.first_success
        else:
            return self._current_game_state.first_success

    @property
    def per_cell(self):
        return self._current_game_state.per_cell

    @per_cell.setter
    def per_cell(self, value):
        self._update_game_state("per_cell", value)

    @property
    def pending_per_cell(self):
        if self.has_pending_game_state():
            return self.pending_game_state.per_cell
        else:
            return self._current_game_state.per_cell

    @property
    def lives(self):
        return self._current_game_state.lives

    @lives.setter
    def lives(self, value):
        self._update_game_state("lives", value)

    @property
    def pending_lives(self):
        if self.has_pending_game_state():
            return self.pending_game_state.lives
        else:
            return self._current_game_state.lives

    @property
    def drag_select(self):
        return self._current_game_state.drag_select

    @drag_select.setter
    def drag_select(self, value):
        self._update_game_state("drag_select", value)

    @property
    def pending_drag_select(self):
        if self.has_pending_game_state():
            return self.pending_game_state.drag_select
        else:
            return self._current_game_state.drag_select

    @property
    def game_mode(self):
        return self._current_game_state.mode

    @game_mode.setter
    def game_mode(self, value):
        self._update_game_state("mode", value)

    @property
    def pending_game_mode(self):
        if self.has_pending_game_state():
            return self.pending_game_state.mode
        else:
            return self._current_game_state.mode

    @property
    def current_game_state(self) -> PerGameState:
        return self._current_game_state

    @property
    def pending_game_state(self) -> PerGameState:
        if self.has_pending_game_state():
            return self._pending_game_state
        else:
            return self._current_game_state

    def has_pending_game_state(self) -> bool:
        return self._pending_game_state is not None

    def _activate_pending_game_state(self) -> None:
        if self.has_pending_game_state():
            self._current_game_state = self._pending_game_state
            self._pending_game_state = None

    @property
    def game_status(self) -> GameState:
        return self._game_status

    @game_status.setter
    def game_status(self, value: GameState):
        self._game_status = value
        if value is GameState.READY and self.has_pending_game_state():
            logger.info(
                "Updating game state on new game from %s to %s",
                self._current_game_state,
                self.pending_game_state,
            )
            self._activate_pending_game_state()

    def reset(self) -> None:
        for name, field in attr.fields_dict(type(self)).items():
            setattr(self, name, field.default)

    # ---------------------------------
    # Methods for instance creation
    # ---------------------------------
    def deepcopy(self) -> "State":
        cls = type(self)
        pending_game_state = None
        if self.has_pending_game_state():
            pending_game_state = self.pending_game_state.copy()
        return cls(
            current_game_state=self._current_game_state.copy(),
            pending_game_state=pending_game_state,
            btn_size=self.btn_size,
            name=self.name,
            styles=self.styles.copy(),
        )

    @classmethod
    def from_opts(cls, game_opts: GameOptsStruct, gui_opts: GUIOptsStruct) -> "State":
        dict_ = {**attr.asdict(game_opts), **attr.asdict(gui_opts)}
        args = {a: v for a, v in dict_.items() if a in attr.fields_dict(cls)}
        args["current_game_state"] = PerGameState.from_structs(game_opts, gui_opts)
        return cls(**args)



# October 2021, Lewis Gaul

__all__ = ("Minefield",)

from typing import Any, List, Mapping

from ...shared.types import CellContents
from ..regular.minefield import RegularMinefieldBase
from .board import Board
from .types import Coord


class Minefield(RegularMinefieldBase[Coord, Board]):
    """A split-cell minefield."""

    # Note: Properties like openings are dependent on the state of the board
    #  (whether cells are split). However, a minefield does still have the
    #  concept of 'minimum clicks without using chording' (i.e. a variant of
    #  3bv) and completed board. We also take 'openings' to mean the openings
    #  on the completed board.

    @classmethod
    def from_json(cls, obj: Mapping[str, Any]) -> "Minefield":
        """
        Create a minefield instance from a JSON encoding.

        :param obj:
            The dictionary obtained from decoding JSON. Must contain the
            following fields: 'x_size', 'y_size', 'mine_coords'.
        :raise ValueError:
            If the dictionary is missing required fields.
        """
        try:
            return cls.from_coords(
                (
                    Coord(x, y, True)
                    for x in range(obj["x_size"])
                    for y in range(obj["y_size"])
                ),
                mine_coords=[Coord(*c, True) for c in obj["mine_coords"]],
                per_cell=obj.get("per_cell", 1),
            )
        except KeyError as e:
            raise ValueError(
                "Missing key in dictionary when trying to create minefield"
            ) from e

    def to_json(self) -> Mapping[str, Any]:
        return dict(
            type="split_cell",
            x_size=self.x_size,
            y_size=self.y_size,
            mine_coords=[(c.x, c.y) for c in self.mine_coords],
            per_cell=self.per_cell,
        )

    def _calc_3bv(self) -> int:
        """Calculate the 3bv of the board."""
        # Number of cells that must be split (contain a safe small cell and a mine).
        big_cells_containing_mines = {
            small.get_big_cell_coord()
            for small in self.completed_board.all_coords
            if small.is_split
        }
        clicks = len(
            [
                big
                for big in big_cells_containing_mines
                if not all(small in self.mine_coords for small in big.split())
            ]
        )
        # Optimally click the openings first to reveal the edge numbers.
        clicks += len(self.openings)
        # Find remaining clicks required to reveal numbers.
        exposed = len({c for opening in self.openings for c in opening})
        clicks += (
            len(self.completed_board.all_coords) - len(set(self.mine_coords)) - exposed
        )
        return clicks

    def _calc_completed_board(self) -> Board:
        """
        Create the completed board with the flags and numbers that should be
        seen upon game completion.
        """
        board = Board(self.x_size, self.y_size)
        # First mark all mines as flags, splitting the containing big cells.
        for coord in set(self.mine_coords):
            if coord not in board:
                big_coord = board.get_coord_at(coord.x, coord.y)
                board.split_coord(big_coord)
            board[coord] = CellContents.Flag(self[coord])
        # Now calculate numbers for all remaining cells.
        for coord in board.all_coords:
            if board[coord] is not CellContents.Unclicked:
                continue
            num = sum(
                self[c]
                for c in board.get_nbrs(coord)
                if type(board[c]) is CellContents.Flag
            )
            board[coord] = CellContents.Num(num)
        return board

    def _find_openings(self) -> List[List[Coord]]:
        board = self.completed_board
        openings = []
        blanks_to_check = {
            c for c in board.all_coords if board[c] is CellContents.Num(0)
        }
        while blanks_to_check:
            orig_coord = blanks_to_check.pop()
            # If the coordinate is part of an opening and hasn't already been
            # considered, start a new opening.
            opening = {orig_coord}  # Coords belonging to the opening
            check = {orig_coord}  # Coords whose neighbours need checking
            while check:
                coord = check.pop()
                nbrs = set(board.get_nbrs(coord))
                check |= {c for c in nbrs - opening if board[c] is CellContents.Num(0)}
                opening |= nbrs
            openings.append(sorted(opening))
            blanks_to_check -= opening
        return openings



# April 2018, Lewis Gaul

__all__ = (
    "BoardBase",
    "GameBase",
    "MinefieldBase",
    "UberController",
    "api",
    "regular",
    "split_cell",
)

from . import api, regular, split_cell
from .board import BoardBase
from .engine import UberController
from .game import GameBase
from .minefield import MinefieldBase
